import csv
from typing import Any, Dict, List, Optional

try:
    import psycopg
    from psycopg import sql
    from psycopg.connection import Connection as PgConnection
    from psycopg.rows import DictRow, dict_row
except ImportError:
    raise ImportError("`psycopg` not installed. Please install using `pip install 'psycopg-binary'`.")

from agno.tools import Toolkit
from agno.utils.log import log_debug, log_error


class PostgresTools(Toolkit):
    def __init__(
        self,
        connection: Optional[PgConnection[DictRow]] = None,
        db_name: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        table_schema: str = "public",
        **kwargs,
    ):
        self._connection: Optional[PgConnection[DictRow]] = connection
        self.db_name: Optional[str] = db_name
        self.user: Optional[str] = user
        self.password: Optional[str] = password
        self.host: Optional[str] = host
        self.port: Optional[int] = port
        self.table_schema: str = table_schema

        tools: List[Any] = [
            self.show_tables,
            self.describe_table,
            self.summarize_table,
            self.inspect_query,
            self.run_query,
            self.export_table_to_path,
        ]

        super().__init__(name="postgres_tools", tools=tools, **kwargs)

    @property
    def connection(self) -> PgConnection[DictRow]:
        """
        Returns the Postgres psycopg connection.
        :return psycopg.connection.Connection: psycopg connection
        """
        if self._connection is None or self._connection.closed:
            log_debug("Establishing new PostgreSQL connection.")
            connection_kwargs: Dict[str, Any] = {"row_factory": dict_row}
            if self.db_name:
                connection_kwargs["dbname"] = self.db_name
            if self.user:
                connection_kwargs["user"] = self.user
            if self.password:
                connection_kwargs["password"] = self.password
            if self.host:
                connection_kwargs["host"] = self.host
            if self.port:
                connection_kwargs["port"] = self.port

            connection_kwargs["options"] = f"-c search_path={self.table_schema}"

            self._connection = psycopg.connect(**connection_kwargs)
            self._connection.read_only = True

        return self._connection

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        """Closes the database connection if it's open."""
        if self._connection and not self._connection.closed:
            log_debug("Closing PostgreSQL connection.")
            self._connection.close()
            self._connection = None

    def _execute_query(self, query: str, params: Optional[tuple] = None) -> str:
        try:
            with self.connection.cursor() as cursor:
                log_debug(f"Running PostgreSQL Query: {query} with Params: {params}")
                cursor.execute(query, params)

                if cursor.description is None:
                    return cursor.statusmessage or "Query executed successfully with no output."

                columns = [desc[0] for desc in cursor.description]
                rows = cursor.fetchall()

                if not rows:
                    return f"Query returned no results.\nColumns: {', '.join(columns)}"

                header = ",".join(columns)
                data_rows = [",".join(map(str, row.values())) for row in rows]
                return f"{header}\n" + "\n".join(data_rows)

        except psycopg.Error as e:
            log_error(f"Database error: {e}")
            if self.connection and not self.connection.closed:
                self.connection.rollback()
            return f"Error executing query: {e}"
        except Exception as e:
            log_error(f"An unexpected error occurred: {e}")
            return f"An unexpected error occurred: {e}"

    def show_tables(self) -> str:
        """Lists all tables in the configured schema."""

        stmt = "SELECT table_name FROM information_schema.tables WHERE table_schema = %s;"
        return self._execute_query(stmt, (self.table_schema,))

    def describe_table(self, table: str) -> str:
        """
        Provides the schema (column name, data type, is nullable) for a given table.

        Args:
            table: The name of the table to describe.

        Returns:
            A string describing the table's columns and data types.
        """
        stmt = """
            SELECT column_name, data_type, is_nullable
            FROM information_schema.columns
            WHERE table_schema = %s AND table_name = %s;
        """
        return self._execute_query(stmt, (self.table_schema, table))

    def summarize_table(self, table: str) -> str:
        """
        Computes and returns key summary statistics for a table's columns.

        Args:
            table: The name of the table to summarize.

        Returns:
            A string containing a summary of the table.
        """
        try:
            with self.connection.cursor() as cursor:
                # First, get column information using a parameterized query
                schema_query = """
                    SELECT column_name, data_type
                    FROM information_schema.columns
                    WHERE table_schema = %s AND table_name = %s;
                """
                cursor.execute(schema_query, (self.table_schema, table))
                columns = cursor.fetchall()
                if not columns:
                    return f"Error: Table '{table}' not found in schema '{self.table_schema}'."

                summary_parts = [f"Summary for table: {table}\n"]
                table_identifier = sql.Identifier(self.table_schema, table)

                for col in columns:
                    col_name, data_type = col["column_name"], col["data_type"]
                    col_identifier = sql.Identifier(col_name)

                    query = None
                    if any(
                        t in data_type for t in ["integer", "numeric", "real", "double precision", "bigint", "smallint"]
                    ):
                        query = sql.SQL("""
                            SELECT
                                COUNT(*) AS total_rows,
                                COUNT({col}) AS non_null_rows,
                                MIN({col}) AS min,
                                MAX({col}) AS max,
                                AVG({col}) AS average,
                                STDDEV({col}) AS std_deviation
                            FROM {tbl};
                        """).format(col=col_identifier, tbl=table_identifier)
                    elif any(t in data_type for t in ["char", "text", "uuid"]):
                        query = sql.SQL("""
                            SELECT
                                COUNT(*) AS total_rows,
                                COUNT({col}) AS non_null_rows,
                                COUNT(DISTINCT {col}) AS unique_values,
                                AVG(LENGTH({col}::text)) as avg_length
                            FROM {tbl};
                        """).format(col=col_identifier, tbl=table_identifier)

                    if query:
                        cursor.execute(query)
                        stats = cursor.fetchone()
                        summary_parts.append(f"\n--- Column: {col_name} (Type: {data_type}) ---")
                        if stats is not None:
                            for key, value in stats.items():
                                val_str = (
                                    f"{value:.2f}" if isinstance(value, float) and value is not None else str(value)
                                )
                                summary_parts.append(f"  {key}: {val_str}")
                        else:
                            summary_parts.append("  No statistics available")

                return "\n".join(summary_parts)

        except psycopg.Error as e:
            return f"Error summarizing table: {e}"

    def inspect_query(self, query: str) -> str:
        """
        Shows the execution plan for a SQL query (using EXPLAIN).

        :param query: The SQL query to inspect.
        :return: The query's execution plan.
        """
        return self._execute_query(f"EXPLAIN {query}")

    def export_table_to_path(self, table: str, path: str) -> str:
        """
        Exports a table's data to a local CSV file.

        :param table: The name of the table to export.
        :param path: The local file path to save the file.
        :return: A confirmation message with the file path.
        """
        log_debug(f"Exporting Table {table} as CSV to local path {path}")

        table_identifier = sql.Identifier(self.table_schema, table)
        stmt = sql.SQL("SELECT * FROM {tbl};").format(tbl=table_identifier)

        try:
            with self.connection.cursor() as cursor:
                cursor.execute(stmt)

                if cursor.description is None:
                    return f"Error: Query returned no description for table '{table}'."

                columns = [desc[0] for desc in cursor.description]

                with open(path, "w", newline="", encoding="utf-8") as f:
                    writer = csv.writer(f)
                    writer.writerow(columns)
                    writer.writerows(row.values() for row in cursor)

            return f"Successfully exported table '{table}' to '{path}'."
        except (psycopg.Error, IOError) as e:
            return f"Error exporting table: {e}"

    def run_query(self, query: str) -> str:
        """
        Runs a read-only SQL query and returns the result.

        :param query: The SQL query to run.
        :return: The query result as a formatted string.
        """
        return self._execute_query(query)



from abc import ABC, abstractmethod
from enum import Enum
from typing import List

from agno.knowledge.document.base import Document


class ChunkingStrategy(ABC):
    """Base class for chunking strategies"""

    @abstractmethod
    def chunk(self, document: Document) -> List[Document]:
        raise NotImplementedError

    def clean_text(self, text: str) -> str:
        """Clean the text by replacing multiple newlines with a single newline"""
        import re

        # Replace multiple newlines with a single newline
        cleaned_text = re.sub(r"\n+", "\n", text)
        # Replace multiple spaces with a single space
        cleaned_text = re.sub(r"\s+", " ", cleaned_text)
        # Replace multiple tabs with a single tab
        cleaned_text = re.sub(r"\t+", "\t", cleaned_text)
        # Replace multiple carriage returns with a single carriage return
        cleaned_text = re.sub(r"\r+", "\r", cleaned_text)
        # Replace multiple form feeds with a single form feed
        cleaned_text = re.sub(r"\f+", "\f", cleaned_text)
        # Replace multiple vertical tabs with a single vertical tab
        cleaned_text = re.sub(r"\v+", "\v", cleaned_text)

        return cleaned_text


class ChunkingStrategyType(str, Enum):
    """Enumeration of available chunking strategies."""

    AGENTIC_CHUNKER = "AgenticChunker"
    DOCUMENT_CHUNKER = "DocumentChunker"
    RECURSIVE_CHUNKER = "RecursiveChunker"
    SEMANTIC_CHUNKER = "SemanticChunker"
    FIXED_SIZE_CHUNKER = "FixedSizeChunker"
    ROW_CHUNKER = "RowChunker"
    MARKDOWN_CHUNKER = "MarkdownChunker"

    @classmethod
    def from_string(cls, strategy_name: str) -> "ChunkingStrategyType":
        """Convert a string to a ChunkingStrategyType."""
        strategy_name_clean = strategy_name.strip()

        # Try exact enum value match first
        for enum_member in cls:
            if enum_member.value == strategy_name_clean:
                return enum_member

        raise ValueError(f"Unsupported chunking strategy: {strategy_name}. Valid options: {[e.value for e in cls]}")


class ChunkingStrategyFactory:
    """Factory for creating chunking strategy instances."""

    @classmethod
    def create_strategy(cls, strategy_type: ChunkingStrategyType, **kwargs) -> ChunkingStrategy:
        """Create an instance of the chunking strategy with the given parameters."""
        strategy_map = {
            ChunkingStrategyType.AGENTIC_CHUNKER: cls._create_agentic_chunking,
            ChunkingStrategyType.DOCUMENT_CHUNKER: cls._create_document_chunking,
            ChunkingStrategyType.RECURSIVE_CHUNKER: cls._create_recursive_chunking,
            ChunkingStrategyType.SEMANTIC_CHUNKER: cls._create_semantic_chunking,
            ChunkingStrategyType.FIXED_SIZE_CHUNKER: cls._create_fixed_chunking,
            ChunkingStrategyType.ROW_CHUNKER: cls._create_row_chunking,
            ChunkingStrategyType.MARKDOWN_CHUNKER: cls._create_markdown_chunking,
        }
        return strategy_map[strategy_type](**kwargs)

    @classmethod
    def _create_agentic_chunking(cls, **kwargs) -> ChunkingStrategy:
        from agno.knowledge.chunking.agentic import AgenticChunking

        # Map chunk_size to max_chunk_size for AgenticChunking
        if "chunk_size" in kwargs and "max_chunk_size" not in kwargs:
            kwargs["max_chunk_size"] = kwargs.pop("chunk_size")
        return AgenticChunking(**kwargs)

    @classmethod
    def _create_document_chunking(cls, **kwargs) -> ChunkingStrategy:
        from agno.knowledge.chunking.document import DocumentChunking

        return DocumentChunking(**kwargs)

    @classmethod
    def _create_recursive_chunking(cls, **kwargs) -> ChunkingStrategy:
        from agno.knowledge.chunking.recursive import RecursiveChunking

        return RecursiveChunking(**kwargs)

    @classmethod
    def _create_semantic_chunking(cls, **kwargs) -> ChunkingStrategy:
        from agno.knowledge.chunking.semantic import SemanticChunking

        return SemanticChunking(**kwargs)

    @classmethod
    def _create_fixed_chunking(cls, **kwargs) -> ChunkingStrategy:
        from agno.knowledge.chunking.fixed import FixedSizeChunking

        return FixedSizeChunking(**kwargs)

    @classmethod
    def _create_row_chunking(cls, **kwargs) -> ChunkingStrategy:
        from agno.knowledge.chunking.row import RowChunking

        # Remove chunk_size if present since RowChunking doesn't use it
        kwargs.pop("chunk_size", None)
        return RowChunking(**kwargs)

    @classmethod
    def _create_markdown_chunking(cls, **kwargs) -> ChunkingStrategy:
        from agno.knowledge.chunking.markdown import MarkdownChunking

        return MarkdownChunking(**kwargs)



import time
from datetime import date, datetime, timedelta, timezone
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union
from uuid import uuid4

from agno.db.async_postgres.schemas import get_table_schema_definition
from agno.db.async_postgres.utils import (
    apply_sorting,
    bulk_upsert_metrics,
    calculate_date_metrics,
    create_schema,
    fetch_all_sessions_data,
    get_dates_to_calculate_metrics_for,
    is_table_available,
    is_valid_table,
)
from agno.db.base import AsyncBaseDb, SessionType
from agno.db.schemas.evals import EvalFilterType, EvalRunRecord, EvalType
from agno.db.schemas.knowledge import KnowledgeRow
from agno.db.schemas.memory import UserMemory
from agno.session import AgentSession, Session, TeamSession, WorkflowSession
from agno.utils.log import log_debug, log_error, log_info, log_warning

try:
    from sqlalchemy import Index, String, UniqueConstraint, func, update
    from sqlalchemy.dialects import postgresql
    from sqlalchemy.ext.asyncio import AsyncEngine, async_sessionmaker, create_async_engine
    from sqlalchemy.schema import Column, MetaData, Table
    from sqlalchemy.sql.expression import select, text
except ImportError:
    raise ImportError("`sqlalchemy` not installed. Please install it using `pip install sqlalchemy`")


class AsyncPostgresDb(AsyncBaseDb):
    def __init__(
        self,
        db_id: Optional[str] = None,
        db_url: Optional[str] = None,
        db_engine: Optional[AsyncEngine] = None,
        db_schema: Optional[str] = None,
        session_table: Optional[str] = None,
        memory_table: Optional[str] = None,
        metrics_table: Optional[str] = None,
        eval_table: Optional[str] = None,
        knowledge_table: Optional[str] = None,
    ):
        """
        Async interface for interacting with a PostgreSQL database.

        The following order is used to determine the database connection:
            1. Use the db_engine if provided
            2. Use the db_url
            3. Raise an error if neither is provided

        Args:
            db_id (Optional[str]): The ID of the database.
            db_url (Optional[str]): The database URL to connect to.
            db_engine (Optional[AsyncEngine]): The SQLAlchemy async database engine to use.
            db_schema (Optional[str]): The database schema to use.
            session_table (Optional[str]): Name of the table to store Agent, Team and Workflow sessions.
            memory_table (Optional[str]): Name of the table to store memories.
            metrics_table (Optional[str]): Name of the table to store metrics.
            eval_table (Optional[str]): Name of the table to store evaluation runs data.
            knowledge_table (Optional[str]): Name of the table to store knowledge content.

        Raises:
            ValueError: If neither db_url nor db_engine is provided.
            ValueError: If none of the tables are provided.
        """
        super().__init__(
            id=db_id,
            session_table=session_table,
            memory_table=memory_table,
            metrics_table=metrics_table,
            eval_table=eval_table,
            knowledge_table=knowledge_table,
        )

        _engine: Optional[AsyncEngine] = db_engine
        if _engine is None and db_url is not None:
            _engine = create_async_engine(db_url)
        if _engine is None:
            raise ValueError("One of db_url or db_engine must be provided")

        self.db_url: Optional[str] = db_url
        self.db_engine: AsyncEngine = _engine
        self.db_schema: str = db_schema if db_schema is not None else "ai"
        self.metadata: MetaData = MetaData()

        # Initialize database session factory
        self.async_session_factory = async_sessionmaker(bind=self.db_engine)

    # -- DB methods --
    async def _create_table(self, table_name: str, table_type: str, db_schema: str) -> Table:
        """
        Create a table with the appropriate schema based on the table type.

        Args:
            table_name (str): Name of the table to create
            table_type (str): Type of table (used to get schema definition)
            db_schema (str): Database schema name

        Returns:
            Table: SQLAlchemy Table object
        """
        try:
            table_schema = get_table_schema_definition(table_type).copy()

            columns: List[Column] = []
            indexes: List[str] = []
            unique_constraints: List[str] = []
            schema_unique_constraints = table_schema.pop("_unique_constraints", [])

            # Get the columns, indexes, and unique constraints from the table schema
            for col_name, col_config in table_schema.items():
                column_args = [col_name, col_config["type"]()]
                column_kwargs = {}
                if col_config.get("primary_key", False):
                    column_kwargs["primary_key"] = True
                if "nullable" in col_config:
                    column_kwargs["nullable"] = col_config["nullable"]
                if col_config.get("index", False):
                    indexes.append(col_name)
                if col_config.get("unique", False):
                    column_kwargs["unique"] = True
                    unique_constraints.append(col_name)
                columns.append(Column(*column_args, **column_kwargs))  # type: ignore

            # Create the table object
            table_metadata = MetaData(schema=db_schema)
            table = Table(table_name, table_metadata, *columns, schema=db_schema)

            # Add multi-column unique constraints with table-specific names
            for constraint in schema_unique_constraints:
                constraint_name = f"{table_name}_{constraint['name']}"
                constraint_columns = constraint["columns"]
                table.append_constraint(UniqueConstraint(*constraint_columns, name=constraint_name))

            # Add indexes to the table definition
            for idx_col in indexes:
                idx_name = f"idx_{table_name}_{idx_col}"
                table.append_constraint(Index(idx_name, idx_col))

            async with self.async_session_factory() as sess, sess.begin():
                await create_schema(session=sess, db_schema=db_schema)

            # Create table
            async with self.db_engine.begin() as conn:
                await conn.run_sync(table.create, checkfirst=True)

            # Create indexes
            for idx in table.indexes:
                try:
                    # Check if index already exists
                    async with self.async_session_factory() as sess:
                        exists_query = text(
                            "SELECT 1 FROM pg_indexes WHERE schemaname = :schema AND indexname = :index_name"
                        )
                        result = await sess.execute(exists_query, {"schema": db_schema, "index_name": idx.name})
                        exists = result.scalar() is not None
                        if exists:
                            log_debug(f"Index {idx.name} already exists in {db_schema}.{table_name}, skipping creation")
                            continue

                    async with self.db_engine.begin() as conn:
                        await conn.run_sync(idx.create)
                    log_debug(f"Created index: {idx.name} for table {db_schema}.{table_name}")

                except Exception as e:
                    log_error(f"Error creating index {idx.name}: {e}")

            log_info(f"Successfully created table {table_name} in schema {db_schema}")
            return table

        except Exception as e:
            log_error(f"Could not create table {db_schema}.{table_name}: {e}")
            raise

    async def _get_table(self, table_type: str) -> Table:
        if table_type == "sessions":
            if not hasattr(self, "session_table"):
                self.session_table = await self._get_or_create_table(
                    table_name=self.session_table_name, table_type="sessions", db_schema=self.db_schema
                )
            return self.session_table

        if table_type == "memories":
            if not hasattr(self, "memory_table"):
                self.memory_table = await self._get_or_create_table(
                    table_name=self.memory_table_name, table_type="memories", db_schema=self.db_schema
                )
            return self.memory_table

        if table_type == "metrics":
            if not hasattr(self, "metrics_table"):
                self.metrics_table = await self._get_or_create_table(
                    table_name=self.metrics_table_name, table_type="metrics", db_schema=self.db_schema
                )
            return self.metrics_table

        if table_type == "evals":
            if not hasattr(self, "eval_table"):
                self.eval_table = await self._get_or_create_table(
                    table_name=self.eval_table_name, table_type="evals", db_schema=self.db_schema
                )
            return self.eval_table

        if table_type == "knowledge":
            if not hasattr(self, "knowledge_table"):
                self.knowledge_table = await self._get_or_create_table(
                    table_name=self.knowledge_table_name, table_type="knowledge", db_schema=self.db_schema
                )
            return self.knowledge_table

        raise ValueError(f"Unknown table type: {table_type}")

    async def _get_or_create_table(self, table_name: str, table_type: str, db_schema: str) -> Table:
        """
        Check if the table exists and is valid, else create it.

        Args:
            table_name (str): Name of the table to get or create
            table_type (str): Type of table (used to get schema definition)
            db_schema (str): Database schema name

        Returns:
            Table: SQLAlchemy Table object representing the schema.
        """

        async with self.async_session_factory() as sess, sess.begin():
            table_is_available = await is_table_available(session=sess, table_name=table_name, db_schema=db_schema)

        if not table_is_available:
            return await self._create_table(table_name=table_name, table_type=table_type, db_schema=db_schema)

        if not await is_valid_table(
            db_engine=self.db_engine,
            table_name=table_name,
            table_type=table_type,
            db_schema=db_schema,
        ):
            raise ValueError(f"Table {db_schema}.{table_name} has an invalid schema")

        try:
            async with self.db_engine.connect() as conn:

                def create_table(connection):
                    return Table(table_name, self.metadata, schema=db_schema, autoload_with=connection)

                table = await conn.run_sync(create_table)
                return table

        except Exception as e:
            log_error(f"Error loading existing table {db_schema}.{table_name}: {e}")
            raise

    # -- Session methods --
    async def delete_session(self, session_id: str) -> bool:
        """
        Delete a session from the database.

        Args:
            session_id (str): ID of the session to delete

        Returns:
            bool: True if the session was deleted, False otherwise.

        Raises:
            Exception: If an error occurs during deletion.
        """
        try:
            table = await self._get_table(table_type="sessions")

            async with self.async_session_factory() as sess, sess.begin():
                delete_stmt = table.delete().where(table.c.session_id == session_id)
                result = await sess.execute(delete_stmt)

                if result.rowcount == 0:
                    log_debug(f"No session found to delete with session_id: {session_id} in table {table.name}")
                    return False

                else:
                    log_debug(f"Successfully deleted session with session_id: {session_id} in table {table.name}")
                    return True

        except Exception as e:
            log_error(f"Error deleting session: {e}")
            return False

    async def delete_sessions(self, session_ids: List[str]) -> None:
        """Delete all given sessions from the database.
        Can handle multiple session types in the same run.

        Args:
            session_ids (List[str]): The IDs of the sessions to delete.

        Raises:
            Exception: If an error occurs during deletion.
        """
        try:
            table = await self._get_table(table_type="sessions")

            async with self.async_session_factory() as sess, sess.begin():
                delete_stmt = table.delete().where(table.c.session_id.in_(session_ids))
                result = await sess.execute(delete_stmt)

            log_debug(f"Successfully deleted {result.rowcount} sessions")

        except Exception as e:
            log_error(f"Error deleting sessions: {e}")

    async def get_session(
        self,
        session_id: str,
        session_type: SessionType,
        user_id: Optional[str] = None,
        deserialize: Optional[bool] = True,
    ) -> Optional[Union[Session, Dict[str, Any]]]:
        """
        Read a session from the database.

        Args:
            session_id (str): ID of the session to read.
            user_id (Optional[str]): User ID to filter by. Defaults to None.
            session_type (Optional[SessionType]): Type of session to read. Defaults to None.
            deserialize (Optional[bool]): Whether to serialize the session. Defaults to True.

        Returns:
            Union[Session, Dict[str, Any], None]:
                - When deserialize=True: Session object
                - When deserialize=False: Session dictionary

        Raises:
            Exception: If an error occurs during retrieval.
        """
        try:
            table = await self._get_table(table_type="sessions")

            async with self.async_session_factory() as sess:
                stmt = select(table).where(table.c.session_id == session_id)

                if user_id is not None:
                    stmt = stmt.where(table.c.user_id == user_id)
                if session_type is not None:
                    session_type_value = session_type.value if isinstance(session_type, SessionType) else session_type
                    stmt = stmt.where(table.c.session_type == session_type_value)
                result = await sess.execute(stmt)
                row = result.fetchone()
                if row is None:
                    return None

                session = dict(row._mapping)

            if not deserialize:
                return session

            if session_type == SessionType.AGENT:
                return AgentSession.from_dict(session)
            elif session_type == SessionType.TEAM:
                return TeamSession.from_dict(session)
            elif session_type == SessionType.WORKFLOW:
                return WorkflowSession.from_dict(session)
            else:
                raise ValueError(f"Invalid session type: {session_type}")

        except Exception as e:
            log_error(f"Exception reading from session table: {e}")
            return None

    async def get_sessions(
        self,
        session_type: Optional[SessionType] = None,
        user_id: Optional[str] = None,
        component_id: Optional[str] = None,
        session_name: Optional[str] = None,
        start_timestamp: Optional[int] = None,
        end_timestamp: Optional[int] = None,
        limit: Optional[int] = None,
        page: Optional[int] = None,
        sort_by: Optional[str] = None,
        sort_order: Optional[str] = None,
        deserialize: Optional[bool] = True,
    ) -> Union[List[Session], Tuple[List[Dict[str, Any]], int]]:
        """
        Get all sessions in the given table. Can filter by user_id and entity_id.

        Args:
            user_id (Optional[str]): The ID of the user to filter by.
            component_id (Optional[str]): The ID of the agent / workflow to filter by.
            start_timestamp (Optional[int]): The start timestamp to filter by.
            end_timestamp (Optional[int]): The end timestamp to filter by.
            session_name (Optional[str]): The name of the session to filter by.
            limit (Optional[int]): The maximum number of sessions to return. Defaults to None.
            page (Optional[int]): The page number to return. Defaults to None.
            sort_by (Optional[str]): The field to sort by. Defaults to None.
            sort_order (Optional[str]): The sort order. Defaults to None.
            deserialize (Optional[bool]): Whether to serialize the sessions. Defaults to True.

        Returns:
            Union[List[Session], Tuple[List[Dict], int]]:
                - When deserialize=True: List of Session objects
                - When deserialize=False: Tuple of (session dictionaries, total count)

        Raises:
            Exception: If an error occurs during retrieval.
        """
        try:
            table = await self._get_table(table_type="sessions")

            async with self.async_session_factory() as sess, sess.begin():
                stmt = select(table)

                # Filtering
                if user_id is not None:
                    stmt = stmt.where(table.c.user_id == user_id)
                if component_id is not None:
                    if session_type == SessionType.AGENT:
                        stmt = stmt.where(table.c.agent_id == component_id)
                    elif session_type == SessionType.TEAM:
                        stmt = stmt.where(table.c.team_id == component_id)
                    elif session_type == SessionType.WORKFLOW:
                        stmt = stmt.where(table.c.workflow_id == component_id)
                if start_timestamp is not None:
                    stmt = stmt.where(table.c.created_at >= start_timestamp)
                if end_timestamp is not None:
                    stmt = stmt.where(table.c.created_at <= end_timestamp)
                if session_name is not None:
                    stmt = stmt.where(
                        func.coalesce(func.json_extract_path_text(table.c.session_data, "session_name"), "").ilike(
                            f"%{session_name}%"
                        )
                    )
                if session_type is not None:
                    session_type_value = session_type.value if isinstance(session_type, SessionType) else session_type
                    stmt = stmt.where(table.c.session_type == session_type_value)

                count_stmt = select(func.count()).select_from(stmt.alias())
                total_count = await sess.scalar(count_stmt) or 0

                # Sorting
                stmt = apply_sorting(stmt, table, sort_by, sort_order)

                # Paginating
                if limit is not None:
                    stmt = stmt.limit(limit)
                    if page is not None:
                        stmt = stmt.offset((page - 1) * limit)

                result = await sess.execute(stmt)
                records = result.fetchall()
                if records is None:
                    return [], 0

                session = [dict(record._mapping) for record in records]
                if not deserialize:
                    return session, total_count

            if session_type == SessionType.AGENT:
                return [AgentSession.from_dict(record) for record in session]  # type: ignore
            elif session_type == SessionType.TEAM:
                return [TeamSession.from_dict(record) for record in session]  # type: ignore
            elif session_type == SessionType.WORKFLOW:
                return [WorkflowSession.from_dict(record) for record in session]  # type: ignore
            else:
                raise ValueError(f"Invalid session type: {session_type}")

        except Exception as e:
            log_error(f"Exception reading from session table: {e}")
            return [] if deserialize else ([], 0)

    async def rename_session(
        self, session_id: str, session_type: SessionType, session_name: str, deserialize: Optional[bool] = True
    ) -> Optional[Union[Session, Dict[str, Any]]]:
        """
        Rename a session in the database.

        Args:
            session_id (str): The ID of the session to rename.
            session_type (SessionType): The type of session to rename.
            session_name (str): The new name for the session.
            deserialize (Optional[bool]): Whether to serialize the session. Defaults to True.

        Returns:
            Optional[Union[Session, Dict[str, Any]]]:
                - When deserialize=True: Session object
                - When deserialize=False: Session dictionary

        Raises:
            Exception: If an error occurs during renaming.
        """
        try:
            table = await self._get_table(table_type="sessions")

            async with self.async_session_factory() as sess, sess.begin():
                stmt = (
                    update(table)
                    .where(table.c.session_id == session_id)
                    .where(table.c.session_type == session_type.value)
                    .values(
                        session_data=func.cast(
                            func.jsonb_set(
                                func.cast(table.c.session_data, postgresql.JSONB),
                                text("'{session_name}'"),
                                func.to_jsonb(session_name),
                            ),
                            postgresql.JSON,
                        )
                    )
                    .returning(*table.c)
                )
                result = await sess.execute(stmt)
                row = result.fetchone()
                if not row:
                    return None

            log_debug(f"Renamed session with id '{session_id}' to '{session_name}'")

            session = dict(row._mapping)
            if not deserialize:
                return session

            # Return the appropriate session type
            if session_type == SessionType.AGENT:
                return AgentSession.from_dict(session)
            elif session_type == SessionType.TEAM:
                return TeamSession.from_dict(session)
            elif session_type == SessionType.WORKFLOW:
                return WorkflowSession.from_dict(session)
            else:
                raise ValueError(f"Invalid session type: {session_type}")

        except Exception as e:
            log_error(f"Exception renaming session: {e}")
            return None

    async def upsert_session(
        self, session: Session, deserialize: Optional[bool] = True
    ) -> Optional[Union[Session, Dict[str, Any]]]:
        """
        Insert or update a session in the database.

        Args:
            session (Session): The session data to upsert.
            deserialize (Optional[bool]): Whether to deserialize the session. Defaults to True.

        Returns:
            Optional[Union[Session, Dict[str, Any]]]:
                - When deserialize=True: Session object
                - When deserialize=False: Session dictionary

        Raises:
            Exception: If an error occurs during upsert.
        """
        try:
            table = await self._get_table(table_type="sessions")
            session_dict = session.to_dict()

            if isinstance(session, AgentSession):
                async with self.async_session_factory() as sess, sess.begin():
                    stmt = postgresql.insert(table).values(
                        session_id=session_dict.get("session_id"),
                        session_type=SessionType.AGENT.value,
                        agent_id=session_dict.get("agent_id"),
                        user_id=session_dict.get("user_id"),
                        runs=session_dict.get("runs"),
                        agent_data=session_dict.get("agent_data"),
                        session_data=session_dict.get("session_data"),
                        summary=session_dict.get("summary"),
                        metadata=session_dict.get("metadata"),
                        created_at=session_dict.get("created_at"),
                        updated_at=session_dict.get("created_at"),
                    )
                    stmt = stmt.on_conflict_do_update(  # type: ignore
                        index_elements=["session_id"],
                        set_=dict(
                            agent_id=session_dict.get("agent_id"),
                            user_id=session_dict.get("user_id"),
                            agent_data=session_dict.get("agent_data"),
                            session_data=session_dict.get("session_data"),
                            summary=session_dict.get("summary"),
                            metadata=session_dict.get("metadata"),
                            runs=session_dict.get("runs"),
                            updated_at=int(time.time()),
                        ),
                    ).returning(table)
                    result = await sess.execute(stmt)
                    row = result.fetchone()
                    if row is None:
                        return None
                    session_dict = dict(row._mapping)

                    log_debug(f"Upserted agent session with id '{session_dict.get('session_id')}'")

                    if not deserialize:
                        return session_dict
                    return AgentSession.from_dict(session_dict)

            elif isinstance(session, TeamSession):
                async with self.async_session_factory() as sess, sess.begin():
                    stmt = postgresql.insert(table).values(
                        session_id=session_dict.get("session_id"),
                        session_type=SessionType.TEAM.value,
                        team_id=session_dict.get("team_id"),
                        user_id=session_dict.get("user_id"),
                        runs=session_dict.get("runs"),
                        team_data=session_dict.get("team_data"),
                        session_data=session_dict.get("session_data"),
                        summary=session_dict.get("summary"),
                        metadata=session_dict.get("metadata"),
                        created_at=session_dict.get("created_at"),
                        updated_at=session_dict.get("created_at"),
                    )
                    stmt = stmt.on_conflict_do_update(  # type: ignore
                        index_elements=["session_id"],
                        set_=dict(
                            team_id=session_dict.get("team_id"),
                            user_id=session_dict.get("user_id"),
                            team_data=session_dict.get("team_data"),
                            session_data=session_dict.get("session_data"),
                            summary=session_dict.get("summary"),
                            metadata=session_dict.get("metadata"),
                            runs=session_dict.get("runs"),
                            updated_at=int(time.time()),
                        ),
                    ).returning(table)
                    result = await sess.execute(stmt)
                    row = result.fetchone()
                    if row is None:
                        return None
                    session_dict = dict(row._mapping)

                    log_debug(f"Upserted team session with id '{session_dict.get('session_id')}'")

                    if not deserialize:
                        return session_dict
                    return TeamSession.from_dict(session_dict)

            elif isinstance(session, WorkflowSession):
                async with self.async_session_factory() as sess, sess.begin():
                    stmt = postgresql.insert(table).values(
                        session_id=session_dict.get("session_id"),
                        session_type=SessionType.WORKFLOW.value,
                        workflow_id=session_dict.get("workflow_id"),
                        user_id=session_dict.get("user_id"),
                        runs=session_dict.get("runs"),
                        workflow_data=session_dict.get("workflow_data"),
                        session_data=session_dict.get("session_data"),
                        summary=session_dict.get("summary"),
                        metadata=session_dict.get("metadata"),
                        created_at=session_dict.get("created_at"),
                        updated_at=session_dict.get("created_at"),
                    )
                    stmt = stmt.on_conflict_do_update(  # type: ignore
                        index_elements=["session_id"],
                        set_=dict(
                            workflow_id=session_dict.get("workflow_id"),
                            user_id=session_dict.get("user_id"),
                            workflow_data=session_dict.get("workflow_data"),
                            session_data=session_dict.get("session_data"),
                            summary=session_dict.get("summary"),
                            metadata=session_dict.get("metadata"),
                            runs=session_dict.get("runs"),
                            updated_at=int(time.time()),
                        ),
                    ).returning(table)
                    result = await sess.execute(stmt)
                    row = result.fetchone()
                    if row is None:
                        return None
                    session_dict = dict(row._mapping)

                    log_debug(f"Upserted workflow session with id '{session_dict.get('session_id')}'")

                    if not deserialize:
                        return session_dict
                    return WorkflowSession.from_dict(session_dict)

            else:
                raise ValueError(f"Invalid session type: {session.session_type}")

        except Exception as e:
            log_error(f"Exception upserting into sessions table: {e}")
            return None

    # -- Memory methods --
    async def delete_user_memory(self, memory_id: str):
        """Delete a user memory from the database.

        Returns:
            bool: True if deletion was successful, False otherwise.

        Raises:
            Exception: If an error occurs during deletion.
        """
        try:
            table = await self._get_table(table_type="memories")

            async with self.async_session_factory() as sess, sess.begin():
                delete_stmt = table.delete().where(table.c.memory_id == memory_id)
                result = await sess.execute(delete_stmt)

                success = result.rowcount > 0
                if success:
                    log_debug(f"Successfully deleted user memory id: {memory_id}")
                else:
                    log_debug(f"No user memory found with id: {memory_id}")

        except Exception as e:
            log_error(f"Error deleting user memory: {e}")

    async def delete_user_memories(self, memory_ids: List[str]) -> None:
        """Delete user memories from the database.

        Args:
            memory_ids (List[str]): The IDs of the memories to delete.

        Raises:
            Exception: If an error occurs during deletion.
        """
        try:
            table = await self._get_table(table_type="memories")

            async with self.async_session_factory() as sess, sess.begin():
                delete_stmt = table.delete().where(table.c.memory_id.in_(memory_ids))
                result = await sess.execute(delete_stmt)

                if result.rowcount == 0:
                    log_debug(f"No user memories found with ids: {memory_ids}")
                else:
                    log_debug(f"Successfully deleted {result.rowcount} user memories")

        except Exception as e:
            log_error(f"Error deleting user memories: {e}")

    async def get_all_memory_topics(self) -> List[str]:
        """Get all memory topics from the database.

        Returns:
            List[str]: List of memory topics.
        """
        try:
            table = await self._get_table(table_type="memories")

            async with self.async_session_factory() as sess, sess.begin():
                stmt = select(func.json_array_elements_text(table.c.topics))
                result = await sess.execute(stmt)
                records = result.fetchall()

                return list(set([record[0] for record in records]))

        except Exception as e:
            log_error(f"Exception reading from memory table: {e}")
            return []

    async def get_user_memory(
        self, memory_id: str, deserialize: Optional[bool] = True
    ) -> Optional[Union[UserMemory, Dict[str, Any]]]:
        """Get a memory from the database.

        Args:
            memory_id (str): The ID of the memory to get.
            deserialize (Optional[bool]): Whether to serialize the memory. Defaults to True.

        Returns:
            Union[UserMemory, Dict[str, Any], None]:
                - When deserialize=True: UserMemory object
                - When deserialize=False: UserMemory dictionary

        Raises:
            Exception: If an error occurs during retrieval.
        """
        try:
            table = await self._get_table(table_type="memories")

            async with self.async_session_factory() as sess, sess.begin():
                stmt = select(table).where(table.c.memory_id == memory_id)

                result = await sess.execute(stmt)
                row = result.fetchone()
                if not row:
                    return None

                memory_raw = dict(row._mapping)
                if not deserialize:
                    return memory_raw

            return UserMemory.from_dict(memory_raw)

        except Exception as e:
            log_error(f"Exception reading from memory table: {e}")
            return None

    async def get_user_memories(
        self,
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        team_id: Optional[str] = None,
        topics: Optional[List[str]] = None,
        search_content: Optional[str] = None,
        limit: Optional[int] = None,
        page: Optional[int] = None,
        sort_by: Optional[str] = None,
        sort_order: Optional[str] = None,
        deserialize: Optional[bool] = True,
    ) -> Union[List[UserMemory], Tuple[List[Dict[str, Any]], int]]:
        """Get all memories from the database as UserMemory objects.

        Args:
            user_id (Optional[str]): The ID of the user to filter by.
            agent_id (Optional[str]): The ID of the agent to filter by.
            team_id (Optional[str]): The ID of the team to filter by.
            topics (Optional[List[str]]): The topics to filter by.
            search_content (Optional[str]): The content to search for.
            limit (Optional[int]): The maximum number of memories to return.
            page (Optional[int]): The page number.
            sort_by (Optional[str]): The column to sort by.
            sort_order (Optional[str]): The order to sort by.
            deserialize (Optional[bool]): Whether to serialize the memories. Defaults to True.

        Returns:
            Union[List[UserMemory], Tuple[List[Dict[str, Any]], int]]:
                - When deserialize=True: List of UserMemory objects
                - When deserialize=False: Tuple of (memory dictionaries, total count)

        Raises:
            Exception: If an error occurs during retrieval.
        """
        try:
            table = await self._get_table(table_type="memories")

            async with self.async_session_factory() as sess, sess.begin():
                stmt = select(table)
                # Filtering
                if user_id is not None:
                    stmt = stmt.where(table.c.user_id == user_id)
                if agent_id is not None:
                    stmt = stmt.where(table.c.agent_id == agent_id)
                if team_id is not None:
                    stmt = stmt.where(table.c.team_id == team_id)
                if topics is not None:
                    for topic in topics:
                        stmt = stmt.where(func.cast(table.c.topics, String).like(f'%"{topic}"%'))
                if search_content is not None:
                    stmt = stmt.where(func.cast(table.c.memory, postgresql.TEXT).ilike(f"%{search_content}%"))

                # Get total count after applying filtering
                count_stmt = select(func.count()).select_from(stmt.alias())
                total_count = await sess.scalar(count_stmt) or 0

                # Sorting
                stmt = apply_sorting(stmt, table, sort_by, sort_order)

                # Paginating
                if limit is not None:
                    stmt = stmt.limit(limit)
                    if page is not None:
                        stmt = stmt.offset((page - 1) * limit)

                result = await sess.execute(stmt)
                records = result.fetchall()
                if not records:
                    return [] if deserialize else ([], 0)

                memories_raw = [dict(record._mapping) for record in records]
                if not deserialize:
                    return memories_raw, total_count

            return [UserMemory.from_dict(record) for record in memories_raw]

        except Exception as e:
            log_error(f"Exception reading from memory table: {e}")
            return [] if deserialize else ([], 0)

    async def clear_memories(self) -> None:
        """Delete all memories from the database.

        Raises:
            Exception: If an error occurs during deletion.
        """
        try:
            table = await self._get_table(table_type="memories")

            async with self.async_session_factory() as sess, sess.begin():
                await sess.execute(table.delete())

        except Exception as e:
            log_warning(f"Exception deleting all memories: {e}")

    async def get_user_memory_stats(
        self, limit: Optional[int] = None, page: Optional[int] = None
    ) -> Tuple[List[Dict[str, Any]], int]:
        """Get user memories stats.

        Args:
            limit (Optional[int]): The maximum number of user stats to return.
            page (Optional[int]): The page number.

        Returns:
            Tuple[List[Dict[str, Any]], int]: A list of dictionaries containing user stats and total count.

        Example:
        (
            [
                {
                    "user_id": "123",
                    "total_memories": 10,
                    "last_memory_updated_at": 1714560000,
                },
            ],
            total_count: 1,
        )
        """
        try:
            table = await self._get_table(table_type="memories")

            async with self.async_session_factory() as sess, sess.begin():
                stmt = (
                    select(
                        table.c.user_id,
                        func.count(table.c.memory_id).label("total_memories"),
                        func.max(table.c.updated_at).label("last_memory_updated_at"),
                    )
                    .where(table.c.user_id.is_not(None))
                    .group_by(table.c.user_id)
                    .order_by(func.max(table.c.updated_at).desc())
                )

                count_stmt = select(func.count()).select_from(stmt.alias())
                total_count = await sess.scalar(count_stmt) or 0

                # Pagination
                if limit is not None:
                    stmt = stmt.limit(limit)
                    if page is not None:
                        stmt = stmt.offset((page - 1) * limit)

                result = await sess.execute(stmt)
                records = result.fetchall()
                if not records:
                    return [], 0

                return [
                    {
                        "user_id": record.user_id,  # type: ignore
                        "total_memories": record.total_memories,
                        "last_memory_updated_at": record.last_memory_updated_at,
                    }
                    for record in records
                ], total_count

        except Exception as e:
            log_error(f"Exception getting user memory stats: {e}")
            return [], 0

    async def upsert_user_memory(
        self, memory: UserMemory, deserialize: Optional[bool] = True
    ) -> Optional[Union[UserMemory, Dict[str, Any]]]:
        """Upsert a user memory in the database.

        Args:
            memory (UserMemory): The user memory to upsert.
            deserialize (Optional[bool]): Whether to serialize the memory. Defaults to True.

        Returns:
            Optional[Union[UserMemory, Dict[str, Any]]]:
                - When deserialize=True: UserMemory object
                - When deserialize=False: UserMemory dictionary

        Raises:
            Exception: If an error occurs during upsert.
        """
        try:
            table = await self._get_table(table_type="memories")

            async with self.async_session_factory() as sess, sess.begin():
                if memory.memory_id is None:
                    memory.memory_id = str(uuid4())

                stmt = postgresql.insert(table).values(
                    memory_id=memory.memory_id,
                    memory=memory.memory,
                    input=memory.input,
                    user_id=memory.user_id,
                    agent_id=memory.agent_id,
                    team_id=memory.team_id,
                    topics=memory.topics,
                    updated_at=int(time.time()),
                )
                stmt = stmt.on_conflict_do_update(  # type: ignore
                    index_elements=["memory_id"],
                    set_=dict(
                        memory=memory.memory,
                        topics=memory.topics,
                        input=memory.input,
                        agent_id=memory.agent_id,
                        team_id=memory.team_id,
                        updated_at=int(time.time()),
                    ),
                ).returning(table)

                result = await sess.execute(stmt)
                row = result.fetchone()
                if row is None:
                    return None

            memory_raw = dict(row._mapping)

            log_debug(f"Upserted user memory with id '{memory.memory_id}'")

            if not memory_raw or not deserialize:
                return memory_raw

            return UserMemory.from_dict(memory_raw)

        except Exception as e:
            log_error(f"Exception upserting user memory: {e}")
            return None

    # -- Metrics methods --
    async def _get_all_sessions_for_metrics_calculation(
        self, start_timestamp: Optional[int] = None, end_timestamp: Optional[int] = None
    ) -> List[Dict[str, Any]]:
        """
        Get all sessions of all types (agent, team, workflow) as raw dictionaries.

         Args:
            start_timestamp (Optional[int]): The start timestamp to filter by. Defaults to None.
            end_timestamp (Optional[int]): The end timestamp to filter by. Defaults to None.

        Returns:
            List[Dict[str, Any]]: List of session dictionaries with session_type field.

        Raises:
            Exception: If an error occurs during retrieval.
        """
        try:
            table = await self._get_table(table_type="sessions")

            stmt = select(
                table.c.user_id,
                table.c.session_data,
                table.c.runs,
                table.c.created_at,
                table.c.session_type,
            )

            if start_timestamp is not None:
                stmt = stmt.where(table.c.created_at >= start_timestamp)
            if end_timestamp is not None:
                stmt = stmt.where(table.c.created_at <= end_timestamp)

            async with self.async_session_factory() as sess:
                result = await sess.execute(stmt)
                records = result.fetchall()

                return [dict(record._mapping) for record in records]

        except Exception as e:
            log_error(f"Exception reading from sessions table: {e}")
            return []

    async def _get_metrics_calculation_starting_date(self, table: Table) -> Optional[date]:
        """Get the first date for which metrics calculation is needed:

        1. If there are metrics records, return the date of the first day without a complete metrics record.
        2. If there are no metrics records, return the date of the first recorded session.
        3. If there are no metrics records and no sessions records, return None.

        Args:
            table (Table): The table to get the starting date for.

        Returns:
            Optional[date]: The starting date for which metrics calculation is needed.
        """
        async with self.async_session_factory() as sess:
            stmt = select(table).order_by(table.c.date.desc()).limit(1)
            result = await sess.execute(stmt)
            row = result.fetchone()

            # 1. Return the date of the first day without a complete metrics record.
            if row is not None:
                if row.completed:
                    return row._mapping["date"] + timedelta(days=1)
                else:
                    return row._mapping["date"]

        # 2. No metrics records. Return the date of the first recorded session.
        first_session, _ = await self.get_sessions(sort_by="created_at", sort_order="asc", limit=1, deserialize=False)

        first_session_date = first_session[0]["created_at"] if first_session else None  # type: ignore[index]

        # 3. No metrics records and no sessions records. Return None.
        if first_session_date is None:
            return None

        return datetime.fromtimestamp(first_session_date, tz=timezone.utc).date()

    async def calculate_metrics(self) -> Optional[list[dict]]:
        """Calculate metrics for all dates without complete metrics.

        Returns:
            Optional[list[dict]]: The calculated metrics.

        Raises:
            Exception: If an error occurs during metrics calculation.
        """
        try:
            table = await self._get_table(table_type="metrics")

            starting_date = await self._get_metrics_calculation_starting_date(table)

            if starting_date is None:
                log_info("No session data found. Won't calculate metrics.")
                return None

            dates_to_process = get_dates_to_calculate_metrics_for(starting_date)
            if not dates_to_process:
                log_info("Metrics already calculated for all relevant dates.")
                return None

            start_timestamp = int(
                datetime.combine(dates_to_process[0], datetime.min.time()).replace(tzinfo=timezone.utc).timestamp()
            )
            end_timestamp = int(
                datetime.combine(dates_to_process[-1] + timedelta(days=1), datetime.min.time())
                .replace(tzinfo=timezone.utc)
                .timestamp()
            )

            sessions = await self._get_all_sessions_for_metrics_calculation(
                start_timestamp=start_timestamp, end_timestamp=end_timestamp
            )

            all_sessions_data = fetch_all_sessions_data(
                sessions=sessions, dates_to_process=dates_to_process, start_timestamp=start_timestamp
            )
            if not all_sessions_data:
                log_info("No new session data found. Won't calculate metrics.")
                return None

            results = []
            metrics_records = []

            for date_to_process in dates_to_process:
                date_key = date_to_process.isoformat()
                sessions_for_date = all_sessions_data.get(date_key, {})

                # Skip dates with no sessions
                if not any(len(sessions) > 0 for sessions in sessions_for_date.values()):
                    continue

                metrics_record = calculate_date_metrics(date_to_process, sessions_for_date)

                metrics_records.append(metrics_record)

            if metrics_records:
                async with self.async_session_factory() as sess, sess.begin():
                    results = await bulk_upsert_metrics(session=sess, table=table, metrics_records=metrics_records)

            log_debug("Updated metrics calculations")

            return results

        except Exception as e:
            log_error(f"Exception refreshing metrics: {e}")
            return None

    async def get_metrics(
        self, starting_date: Optional[date] = None, ending_date: Optional[date] = None
    ) -> Tuple[List[dict], Optional[int]]:
        """Get all metrics matching the given date range.

        Args:
            starting_date (Optional[date]): The starting date to filter metrics by.
            ending_date (Optional[date]): The ending date to filter metrics by.

        Returns:
            Tuple[List[dict], Optional[int]]: A tuple containing the metrics and the timestamp of the latest update.

        Raises:
            Exception: If an error occurs during retrieval.
        """
        try:
            table = await self._get_table(table_type="metrics")

            async with self.async_session_factory() as sess, sess.begin():
                stmt = select(table)
                if starting_date:
                    stmt = stmt.where(table.c.date >= starting_date)
                if ending_date:
                    stmt = stmt.where(table.c.date <= ending_date)
                result = await sess.execute(stmt)
                records = result.fetchall()
                if not records:
                    return [], None

                # Get the latest updated_at
                latest_stmt = select(func.max(table.c.updated_at))
                latest_result = await sess.execute(latest_stmt)
                latest_updated_at = latest_result.scalar()

            return [dict(row._mapping) for row in records], latest_updated_at

        except Exception as e:
            log_warning(f"Exception getting metrics: {e}")
            return [], None

    # -- Knowledge methods --
    async def delete_knowledge_content(self, id: str):
        """Delete a knowledge row from the database.

        Args:
            id (str): The ID of the knowledge row to delete.
        """
        table = await self._get_table(table_type="knowledge")

        try:
            async with self.async_session_factory() as sess, sess.begin():
                stmt = table.delete().where(table.c.id == id)
                await sess.execute(stmt)

        except Exception as e:
            log_error(f"Exception deleting knowledge content: {e}")

    async def get_knowledge_content(self, id: str) -> Optional[KnowledgeRow]:
        """Get a knowledge row from the database.

        Args:
            id (str): The ID of the knowledge row to get.

        Returns:
            Optional[KnowledgeRow]: The knowledge row, or None if it doesn't exist.
        """
        table = await self._get_table(table_type="knowledge")

        try:
            async with self.async_session_factory() as sess, sess.begin():
                stmt = select(table).where(table.c.id == id)
                result = await sess.execute(stmt)
                row = result.fetchone()
                if row is None:
                    return None

                return KnowledgeRow.model_validate(row._mapping)

        except Exception as e:
            log_error(f"Exception getting knowledge content: {e}")
            return None

    async def get_knowledge_contents(
        self,
        limit: Optional[int] = None,
        page: Optional[int] = None,
        sort_by: Optional[str] = None,
        sort_order: Optional[str] = None,
    ) -> Tuple[List[KnowledgeRow], int]:
        """Get all knowledge contents from the database.

        Args:
            limit (Optional[int]): The maximum number of knowledge contents to return.
            page (Optional[int]): The page number.
            sort_by (Optional[str]): The column to sort by.
            sort_order (Optional[str]): The order to sort by.

        Returns:
            List[KnowledgeRow]: The knowledge contents.

        Raises:
            Exception: If an error occurs during retrieval.
        """
        table = await self._get_table(table_type="knowledge")

        try:
            async with self.async_session_factory() as sess, sess.begin():
                stmt = select(table)

                # Apply sorting
                if sort_by is not None:
                    stmt = stmt.order_by(getattr(table.c, sort_by) * (1 if sort_order == "asc" else -1))

                # Get total count before applying limit and pagination
                count_stmt = select(func.count()).select_from(stmt.alias())
                total_count = await sess.scalar(count_stmt) or 0

                # Apply pagination after count
                if limit is not None:
                    stmt = stmt.limit(limit)
                    if page is not None:
                        stmt = stmt.offset((page - 1) * limit)

                result = await sess.execute(stmt)
                records = result.fetchall()
                return [KnowledgeRow.model_validate(record._mapping) for record in records], total_count

        except Exception as e:
            log_error(f"Exception getting knowledge contents: {e}")
            return [], 0

    async def upsert_knowledge_content(self, knowledge_row: KnowledgeRow):
        """Upsert knowledge content in the database.

        Args:
            knowledge_row (KnowledgeRow): The knowledge row to upsert.

        Returns:
            Optional[KnowledgeRow]: The upserted knowledge row, or None if the operation fails.
        """
        try:
            table = await self._get_table(table_type="knowledge")
            async with self.async_session_factory() as sess, sess.begin():
                # Get the actual table columns to avoid "unconsumed column names" error
                table_columns = set(table.columns.keys())

                # Only include fields that exist in the table and are not None
                insert_data = {}
                update_fields = {}

                # Map of KnowledgeRow fields to table columns
                field_mapping = {
                    "id": "id",
                    "name": "name",
                    "description": "description",
                    "metadata": "metadata",
                    "type": "type",
                    "size": "size",
                    "linked_to": "linked_to",
                    "access_count": "access_count",
                    "status": "status",
                    "status_message": "status_message",
                    "created_at": "created_at",
                    "updated_at": "updated_at",
                    "external_id": "external_id",
                }

                # Build insert and update data only for fields that exist in the table
                for model_field, table_column in field_mapping.items():
                    if table_column in table_columns:
                        value = getattr(knowledge_row, model_field, None)
                        if value is not None:
                            insert_data[table_column] = value
                            # Don't include ID in update_fields since it's the primary key
                            if table_column != "id":
                                update_fields[table_column] = value

                # Ensure id is always included for the insert
                if "id" in table_columns and knowledge_row.id:
                    insert_data["id"] = knowledge_row.id

                # Handle case where update_fields is empty (all fields are None or don't exist in table)
                if not update_fields:
                    # If we have insert_data, just do an insert without conflict resolution
                    if insert_data:
                        stmt = postgresql.insert(table).values(insert_data)
                        await sess.execute(stmt)
                    else:
                        # If we have no data at all, this is an error
                        log_error("No valid fields found for knowledge row upsert")
                        return None
                else:
                    # Normal upsert with conflict resolution
                    stmt = (
                        postgresql.insert(table)
                        .values(insert_data)
                        .on_conflict_do_update(index_elements=["id"], set_=update_fields)
                    )
                    await sess.execute(stmt)

            log_debug(f"Upserted knowledge row with id '{knowledge_row.id}'")

            return knowledge_row

        except Exception as e:
            log_error(f"Error upserting knowledge row: {e}")
            return None

    # -- Eval methods --
    async def create_eval_run(self, eval_run: EvalRunRecord) -> Optional[EvalRunRecord]:
        """Create an EvalRunRecord in the database.

        Args:
            eval_run (EvalRunRecord): The eval run to create.

        Returns:
            Optional[EvalRunRecord]: The created eval run, or None if the operation fails.

        Raises:
            Exception: If an error occurs during creation.
        """
        try:
            table = await self._get_table(table_type="evals")

            async with self.async_session_factory() as sess, sess.begin():
                current_time = int(time.time())
                stmt = postgresql.insert(table).values(
                    {"created_at": current_time, "updated_at": current_time, **eval_run.model_dump()}
                )
                await sess.execute(stmt)

            log_debug(f"Created eval run with id '{eval_run.run_id}'")

            return eval_run

        except Exception as e:
            log_error(f"Error creating eval run: {e}")
            return None

    async def delete_eval_run(self, eval_run_id: str) -> None:
        """Delete an eval run from the database.

        Args:
            eval_run_id (str): The ID of the eval run to delete.
        """
        try:
            table = await self._get_table(table_type="evals")

            async with self.async_session_factory() as sess, sess.begin():
                stmt = table.delete().where(table.c.run_id == eval_run_id)
                result = await sess.execute(stmt)

                if result.rowcount == 0:
                    log_warning(f"No eval run found with ID: {eval_run_id}")
                else:
                    log_debug(f"Deleted eval run with ID: {eval_run_id}")

        except Exception as e:
            log_error(f"Error deleting eval run {eval_run_id}: {e}")

    async def delete_eval_runs(self, eval_run_ids: List[str]) -> None:
        """Delete multiple eval runs from the database.

        Args:
            eval_run_ids (List[str]): List of eval run IDs to delete.
        """
        try:
            table = await self._get_table(table_type="evals")

            async with self.async_session_factory() as sess, sess.begin():
                stmt = table.delete().where(table.c.run_id.in_(eval_run_ids))
                result = await sess.execute(stmt)

                if result.rowcount == 0:
                    log_warning(f"No eval runs found with IDs: {eval_run_ids}")
                else:
                    log_debug(f"Deleted {result.rowcount} eval runs")

        except Exception as e:
            log_error(f"Error deleting eval runs {eval_run_ids}: {e}")

    async def get_eval_run(
        self, eval_run_id: str, deserialize: Optional[bool] = True
    ) -> Optional[Union[EvalRunRecord, Dict[str, Any]]]:
        """Get an eval run from the database.

        Args:
            eval_run_id (str): The ID of the eval run to get.
            deserialize (Optional[bool]): Whether to serialize the eval run. Defaults to True.

        Returns:
            Optional[Union[EvalRunRecord, Dict[str, Any]]]:
                - When deserialize=True: EvalRunRecord object
                - When deserialize=False: EvalRun dictionary

        Raises:
            Exception: If an error occurs during retrieval.
        """
        try:
            table = await self._get_table(table_type="evals")

            async with self.async_session_factory() as sess, sess.begin():
                stmt = select(table).where(table.c.run_id == eval_run_id)
                result = await sess.execute(stmt)
                row = result.fetchone()
                if row is None:
                    return None

                eval_run_raw = dict(row._mapping)
                if not deserialize:
                    return eval_run_raw

                return EvalRunRecord.model_validate(eval_run_raw)

        except Exception as e:
            log_error(f"Exception getting eval run {eval_run_id}: {e}")
            return None

    async def get_eval_runs(
        self,
        limit: Optional[int] = None,
        page: Optional[int] = None,
        sort_by: Optional[str] = None,
        sort_order: Optional[str] = None,
        agent_id: Optional[str] = None,
        team_id: Optional[str] = None,
        workflow_id: Optional[str] = None,
        model_id: Optional[str] = None,
        filter_type: Optional[EvalFilterType] = None,
        eval_type: Optional[List[EvalType]] = None,
        deserialize: Optional[bool] = True,
    ) -> Union[List[EvalRunRecord], Tuple[List[Dict[str, Any]], int]]:
        """Get all eval runs from the database.

        Args:
            limit (Optional[int]): The maximum number of eval runs to return.
            page (Optional[int]): The page number.
            sort_by (Optional[str]): The column to sort by.
            sort_order (Optional[str]): The order to sort by.
            agent_id (Optional[str]): The ID of the agent to filter by.
            team_id (Optional[str]): The ID of the team to filter by.
            workflow_id (Optional[str]): The ID of the workflow to filter by.
            model_id (Optional[str]): The ID of the model to filter by.
            eval_type (Optional[List[EvalType]]): The type(s) of eval to filter by.
            filter_type (Optional[EvalFilterType]): Filter by component type (agent, team, workflow).
            deserialize (Optional[bool]): Whether to serialize the eval runs. Defaults to True.

        Returns:
            Union[List[EvalRunRecord], Tuple[List[Dict[str, Any]], int]]:
                - When deserialize=True: List of EvalRunRecord objects
                - When deserialize=False: List of dictionaries

        Raises:
            Exception: If an error occurs during retrieval.
        """
        try:
            table = await self._get_table(table_type="evals")

            async with self.async_session_factory() as sess, sess.begin():
                stmt = select(table)

                # Filtering
                if agent_id is not None:
                    stmt = stmt.where(table.c.agent_id == agent_id)
                if team_id is not None:
                    stmt = stmt.where(table.c.team_id == team_id)
                if workflow_id is not None:
                    stmt = stmt.where(table.c.workflow_id == workflow_id)
                if model_id is not None:
                    stmt = stmt.where(table.c.model_id == model_id)
                if eval_type is not None and len(eval_type) > 0:
                    stmt = stmt.where(table.c.eval_type.in_(eval_type))
                if filter_type is not None:
                    if filter_type == EvalFilterType.AGENT:
                        stmt = stmt.where(table.c.agent_id.is_not(None))
                    elif filter_type == EvalFilterType.TEAM:
                        stmt = stmt.where(table.c.team_id.is_not(None))
                    elif filter_type == EvalFilterType.WORKFLOW:
                        stmt = stmt.where(table.c.workflow_id.is_not(None))

                # Get total count after applying filtering
                count_stmt = select(func.count()).select_from(stmt.alias())
                total_count = await sess.scalar(count_stmt) or 0

                # Sorting
                if sort_by is None:
                    stmt = stmt.order_by(table.c.created_at.desc())
                else:
                    stmt = apply_sorting(stmt, table, sort_by, sort_order)

                # Paginating
                if limit is not None:
                    stmt = stmt.limit(limit)
                    if page is not None:
                        stmt = stmt.offset((page - 1) * limit)

                result = await sess.execute(stmt)
                records = result.fetchall()
                if not records:
                    return [] if deserialize else ([], 0)

                eval_runs_raw = [dict(row._mapping) for row in records]
                if not deserialize:
                    return eval_runs_raw, total_count

                return [EvalRunRecord.model_validate(row) for row in eval_runs_raw]

        except Exception as e:
            log_error(f"Exception getting eval runs: {e}")
            return [] if deserialize else ([], 0)

    async def rename_eval_run(
        self, eval_run_id: str, name: str, deserialize: Optional[bool] = True
    ) -> Optional[Union[EvalRunRecord, Dict[str, Any]]]:
        """Upsert the name of an eval run in the database, returning raw dictionary.

        Args:
            eval_run_id (str): The ID of the eval run to update.
            name (str): The new name of the eval run.

        Returns:
            Optional[Dict[str, Any]]: The updated eval run, or None if the operation fails.

        Raises:
            Exception: If an error occurs during update.
        """
        try:
            table = await self._get_table(table_type="evals")
            async with self.async_session_factory() as sess, sess.begin():
                stmt = (
                    table.update().where(table.c.run_id == eval_run_id).values(name=name, updated_at=int(time.time()))
                )
                await sess.execute(stmt)

            eval_run_raw = await self.get_eval_run(eval_run_id=eval_run_id, deserialize=deserialize)
            if not eval_run_raw or not deserialize:
                return eval_run_raw

            return EvalRunRecord.model_validate(eval_run_raw)

        except Exception as e:
            log_error(f"Error upserting eval run name {eval_run_id}: {e}")
            return None

    # -- Migrations --

    async def migrate_table_from_v1_to_v2(self, v1_db_schema: str, v1_table_name: str, v1_table_type: str):
        """Migrate all content in the given table to the right v2 table"""

        from agno.db.migrations.v1_to_v2 import (
            get_all_table_content,
            parse_agent_sessions,
            parse_memories,
            parse_team_sessions,
            parse_workflow_sessions,
        )

        # Get all content from the old table
        old_content: list[dict[str, Any]] = get_all_table_content(
            db=self,
            db_schema=v1_db_schema,
            table_name=v1_table_name,
        )
        if not old_content:
            log_info(f"No content to migrate from table {v1_table_name}")
            return

        # Parse the content into the new format
        memories: List[UserMemory] = []
        sessions: Sequence[Union[AgentSession, TeamSession, WorkflowSession]] = []
        if v1_table_type == "agent_sessions":
            sessions = parse_agent_sessions(old_content)
        elif v1_table_type == "team_sessions":
            sessions = parse_team_sessions(old_content)
        elif v1_table_type == "workflow_sessions":
            sessions = parse_workflow_sessions(old_content)
        elif v1_table_type == "memories":
            memories = parse_memories(old_content)
        else:
            raise ValueError(f"Invalid table type: {v1_table_type}")

        # Insert the new content into the new table
        if v1_table_type == "agent_sessions":
            for session in sessions:
                await self.upsert_session(session)
            log_info(f"Migrated {len(sessions)} Agent sessions to table: {self.session_table}")

        elif v1_table_type == "team_sessions":
            for session in sessions:
                await self.upsert_session(session)
            log_info(f"Migrated {len(sessions)} Team sessions to table: {self.session_table}")

        elif v1_table_type == "workflow_sessions":
            for session in sessions:
                await self.upsert_session(session)
            log_info(f"Migrated {len(sessions)} Workflow sessions to table: {self.session_table}")

        elif v1_table_type == "memories":
            for memory in memories:
                await self.upsert_user_memory(memory)
            log_info(f"Migrated {len(memories)} memories to table: {self.memory_table}")



from enum import Enum
from typing import Any

from pydantic import BaseModel


class ContentType(str, Enum):
    """Enum for content types supported by knowledge readers."""

    # Generic types
    FILE = "file"
    URL = "url"
    TEXT = "text"
    TOPIC = "topic"
    YOUTUBE = "youtube"

    # Document file extensions
    PDF = ".pdf"
    TXT = ".txt"
    MARKDOWN = ".md"
    DOCX = ".docx"
    DOC = ".doc"
    JSON = ".json"

    # Spreadsheet file extensions
    CSV = ".csv"
    XLSX = ".xlsx"
    XLS = ".xls"


def get_content_type_enum(content_type_str: str) -> ContentType:
    """Convert a content type string to ContentType enum."""
    return ContentType(content_type_str)


class KnowledgeFilter(BaseModel):
    key: str
    value: Any



from typing import Any, Dict, Optional

from pydantic import BaseModel, Field

from agno.api.schemas.utils import get_sdk_version


class OSLaunch(BaseModel):
    """Data sent to API to create an OS Launch"""

    os_id: Optional[str] = None
    data: Optional[Dict[Any, Any]] = None

    sdk_version: str = Field(default_factory=get_sdk_version)



import logging
from functools import lru_cache
from os import getenv
from typing import Any, Literal, Optional

from rich.logging import RichHandler
from rich.text import Text

LOGGER_NAME = "agno"
TEAM_LOGGER_NAME = f"{LOGGER_NAME}-team"
WORKFLOW_LOGGER_NAME = f"{LOGGER_NAME}-workflow"

# Define custom styles for different log sources
LOG_STYLES = {
    "agent": {
        "debug": "green",
        "info": "blue",
    },
    "team": {
        "debug": "magenta",
        "info": "steel_blue1",
    },
    "workflow": {
        "debug": "sandy_brown",
        "info": "orange3",
    },
}


class ColoredRichHandler(RichHandler):
    def __init__(self, *args, source_type: Optional[str] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.source_type = source_type

    def get_level_text(self, record: logging.LogRecord) -> Text:
        # Return empty Text if message is empty
        if not record.msg:
            return Text("")

        level_name = record.levelname.lower()
        if self.source_type and self.source_type in LOG_STYLES:
            if level_name in LOG_STYLES[self.source_type]:
                color = LOG_STYLES[self.source_type][level_name]
                return Text(record.levelname, style=color)
        else:
            if level_name in LOG_STYLES["agent"]:
                color = LOG_STYLES["agent"][level_name]
                return Text(record.levelname, style=color)
        return super().get_level_text(record)


class AgnoLogger(logging.Logger):
    def __init__(self, name: str, level: int = logging.NOTSET):
        super().__init__(name, level)

    def debug(self, msg: str, center: bool = False, symbol: str = "*", *args, **kwargs):  # type: ignore
        if center:
            msg = center_header(str(msg), symbol)
        super().debug(msg, *args, **kwargs)

    def info(self, msg: str, center: bool = False, symbol: str = "*", *args, **kwargs):  # type: ignore
        if center:
            msg = center_header(str(msg), symbol)
        super().info(msg, *args, **kwargs)


def build_logger(logger_name: str, source_type: Optional[str] = None) -> Any:
    # If a logger with the name "agno.{source_type}" is already set, we want to use that one
    _logger = logging.getLogger(f"agno.{logger_name}")
    if _logger.handlers or _logger.level != logging.NOTSET:
        return _logger

    # Set the custom logger class as the default for this logger
    logging.setLoggerClass(AgnoLogger)

    # Create logger with custom class
    _logger = logging.getLogger(logger_name)

    # Reset logger class to default to avoid affecting other loggers
    logging.setLoggerClass(logging.Logger)

    # https://rich.readthedocs.io/en/latest/reference/logging.html#rich.logging.RichHandler
    # https://rich.readthedocs.io/en/latest/logging.html#handle-exceptions
    rich_handler = ColoredRichHandler(
        show_time=False,
        rich_tracebacks=False,
        show_path=True if getenv("AGNO_API_RUNTIME") == "dev" else False,
        tracebacks_show_locals=False,
        source_type=source_type or "agent",
    )
    rich_handler.setFormatter(
        logging.Formatter(
            fmt="%(message)s",
            datefmt="[%X]",
        )
    )

    _logger.addHandler(rich_handler)
    _logger.setLevel(logging.INFO)
    _logger.propagate = False
    return _logger


agent_logger: AgnoLogger = build_logger(LOGGER_NAME, source_type="agent")
team_logger: AgnoLogger = build_logger(TEAM_LOGGER_NAME, source_type="team")
workflow_logger: AgnoLogger = build_logger(WORKFLOW_LOGGER_NAME, source_type="workflow")

# Set the default logger to the agent logger
logger: AgnoLogger = agent_logger

debug_on: bool = False
debug_level: Literal[1, 2] = 1


def set_log_level_to_debug(source_type: Optional[str] = None, level: Literal[1, 2] = 1):
    if source_type is None:
        use_agent_logger()

    _logger = logging.getLogger(LOGGER_NAME if source_type is None else f"{LOGGER_NAME}-{source_type}")
    _logger.setLevel(logging.DEBUG)

    global debug_on
    debug_on = True

    global debug_level
    debug_level = level


def set_log_level_to_info(source_type: Optional[str] = None):
    _logger = logging.getLogger(LOGGER_NAME if source_type is None else f"{LOGGER_NAME}-{source_type}")
    _logger.setLevel(logging.INFO)

    global debug_on
    debug_on = False


def set_log_level_to_warning(source_type: Optional[str] = None):
    _logger = logging.getLogger(LOGGER_NAME if source_type is None else f"{LOGGER_NAME}-{source_type}")
    _logger.setLevel(logging.WARNING)

    global debug_on
    debug_on = False


def set_log_level_to_error(source_type: Optional[str] = None):
    _logger = logging.getLogger(LOGGER_NAME if source_type is None else f"{LOGGER_NAME}-{source_type}")
    _logger.setLevel(logging.ERROR)

    global debug_on
    debug_on = False


def center_header(message: str, symbol: str = "*") -> str:
    try:
        import shutil

        terminal_width = shutil.get_terminal_size().columns
    except Exception:
        terminal_width = 80  # fallback width

    header = f" {message} "
    return f"{header.center(terminal_width - 20, symbol)}"


def use_team_logger():
    """Switch the default logger to use team_logger"""
    global logger
    logger = team_logger


def use_agent_logger():
    """Switch the default logger to use the default agent logger"""
    global logger
    logger = agent_logger


def use_workflow_logger():
    """Switch the default logger to use workflow_logger"""
    global logger
    logger = workflow_logger


@lru_cache(maxsize=128)
def _using_default_logger(logger_instance: Any) -> bool:
    """Return True if the currently active logger is our default AgnoLogger"""
    return isinstance(logger_instance, AgnoLogger)


def log_debug(msg, center: bool = False, symbol: str = "*", log_level: Literal[1, 2] = 1, *args, **kwargs):
    global logger
    global debug_on
    global debug_level

    if debug_on:
        if debug_level >= log_level:
            if _using_default_logger(logger):
                logger.debug(msg, center, symbol, *args, **kwargs)
            else:
                logger.debug(msg, *args, **kwargs)


def log_info(msg, center: bool = False, symbol: str = "*", *args, **kwargs):
    global logger
    if _using_default_logger(logger):
        logger.info(msg, center, symbol, *args, **kwargs)
    else:
        logger.info(msg, *args, **kwargs)


def log_warning(msg, *args, **kwargs):
    global logger
    logger.warning(msg, *args, **kwargs)


def log_error(msg, *args, **kwargs):
    global logger
    logger.error(msg, *args, **kwargs)


def log_exception(msg, *args, **kwargs):
    global logger
    logger.exception(msg, *args, **kwargs)


def configure_agno_logging(
    custom_default_logger: Optional[Any] = None,
    custom_agent_logger: Optional[Any] = None,
    custom_team_logger: Optional[Any] = None,
    custom_workflow_logger: Optional[Any] = None,
) -> None:
    """
    Util to set custom loggers. These will be used everywhere across the Agno library.

    Args:
        custom_default_logger: Default logger to use (overrides agent_logger for default)
        custom_agent_logger: Custom logger for agent operations
        custom_team_logger: Custom logger for team operations
        custom_workflow_logger: Custom logger for workflow operations
    """
    if custom_default_logger is not None:
        global logger
        logger = custom_default_logger

    if custom_agent_logger is not None:
        global agent_logger
        agent_logger = custom_agent_logger

    if custom_team_logger is not None:
        global team_logger
        team_logger = custom_team_logger

    if custom_workflow_logger is not None:
        global workflow_logger
        workflow_logger = custom_workflow_logger



from os import getenv
from pathlib import Path
from typing import Any, Iterable, Iterator, List, Optional, Tuple, Union
from urllib.parse import urlparse
from uuid import uuid4

from agno.agent import Agent
from agno.media import Image, Video
from agno.team.team import Team
from agno.tools import Toolkit
from agno.tools.function import ToolResult
from agno.utils.log import logger

try:
    import replicate
    from replicate.helpers import FileOutput
except ImportError:
    raise ImportError("`replicate` not installed. Please install using `pip install replicate`.")


class ReplicateTools(Toolkit):
    def __init__(
        self,
        api_key: Optional[str] = None,
        model: str = "minimax/video-01",
        enable_generate_media: bool = True,
        all: bool = False,
        **kwargs,
    ):
        self.api_key = api_key or getenv("REPLICATE_API_KEY")
        if not self.api_key:
            logger.error("REPLICATE_API_KEY not set. Please set the REPLICATE_API_KEY environment variable.")
        self.model = model

        tools: List[Any] = []
        if all or enable_generate_media:
            tools.append(self.generate_media)

        super().__init__(name="replicate_toolkit", tools=tools, **kwargs)

    def generate_media(self, agent: Union[Agent, Team], prompt: str) -> ToolResult:
        """
        Use this function to generate an image or a video using a replicate model.
        Args:
            prompt (str): A text description of the content.
        Returns:
            ToolResult: A ToolResult containing the generated media or error message.
        """
        if not self.api_key:
            logger.error("API key is not set. Please provide a valid API key.")
            return ToolResult(content="API key is not set.")

        try:
            outputs = replicate.run(ref=self.model, input={"prompt": prompt})
            if isinstance(outputs, FileOutput):
                outputs = [outputs]
            elif isinstance(outputs, (Iterable, Iterator)) and not isinstance(outputs, str):
                outputs = list(outputs)
            else:
                logger.error(f"Unexpected output type: {type(outputs)}")
                return ToolResult(content=f"Unexpected output type: {type(outputs)}")

            images = []
            videos = []
            results = []

            for output in outputs:
                if not isinstance(output, FileOutput):
                    logger.error(f"Unexpected output type: {type(output)}")
                    return ToolResult(content=f"Unexpected output type: {type(output)}")

                result_msg, media_artifact = self._parse_output(output)
                results.append(result_msg)

                if isinstance(media_artifact, Image):
                    images.append(media_artifact)
                elif isinstance(media_artifact, Video):
                    videos.append(media_artifact)

            content = "\n".join(results)
            return ToolResult(
                content=content,
                images=images if images else None,
                videos=videos if videos else None,
            )
        except Exception as e:
            logger.error(f"Failed to generate media: {e}")
            return ToolResult(content=f"Error: {e}")

    def _parse_output(self, output: FileOutput) -> Tuple[str, Union[Image, Video]]:
        """
        Parse the outputs from the replicate model.
        """
        # Parse the URL to extract the file extension
        parsed_url = urlparse(output.url)
        path = parsed_url.path
        ext = Path(path).suffix.lower()

        # Define supported extensions
        image_extensions = {".jpg", ".jpeg", ".png", ".gif", ".bmp", ".tiff", ".webp"}
        video_extensions = {".mp4", ".mov", ".avi", ".mkv", ".flv", ".wmv", ".webm"}

        media_id = str(uuid4())
        artifact: Union[Image, Video]
        media_type: str

        if ext in image_extensions:
            artifact = Image(id=media_id, url=output.url)
            media_type = "image"
        elif ext in video_extensions:
            artifact = Video(id=media_id, url=output.url)
            media_type = "video"
        else:
            logger.error(f"Unsupported media type with extension '{ext}' for URL: {output.url}")
            raise ValueError(f"Unsupported media type with extension '{ext}'.")

        return f"{media_type.capitalize()} generated successfully at {output.url}", artifact



from typing import Any, Dict, Final, List, Optional, Union

try:
    from surrealdb import (
        AsyncHttpSurrealConnection,
        AsyncWsSurrealConnection,
        BlockingHttpSurrealConnection,
        BlockingWsSurrealConnection,
    )
except ImportError as e:
    msg = "The `surrealdb` package is not installed. Please install it via `pip install surrealdb`."
    raise ImportError(msg) from e

from agno.knowledge.document import Document
from agno.knowledge.embedder import Embedder
from agno.utils.log import log_debug, log_error, log_info
from agno.vectordb.base import VectorDb
from agno.vectordb.distance import Distance


class SurrealDb(VectorDb):
    """SurrealDB Vector Database implementation supporting both sync and async operations."""

    # SQL Query Constants
    CREATE_TABLE_QUERY: Final[str] = """
        DEFINE TABLE IF NOT EXISTS {collection} SCHEMAFUL;
        DEFINE FIELD IF NOT EXISTS content ON {collection} TYPE string;
        DEFINE FIELD IF NOT EXISTS embedding ON {collection} TYPE array<float>;
        DEFINE FIELD IF NOT EXISTS meta_data ON {collection} FLEXIBLE TYPE object;
        DEFINE INDEX IF NOT EXISTS vector_idx ON {collection} FIELDS embedding HNSW DIMENSION {dimensions} DIST {distance};
    """

    DOC_EXISTS_QUERY: Final[str] = """
        SELECT * FROM {collection}
        WHERE content = $content
        LIMIT 1
    """

    NAME_EXISTS_QUERY: Final[str] = """
        SELECT * FROM {collection}
        WHERE meta_data.name = $name
        LIMIT 1
    """

    ID_EXISTS_QUERY: Final[str] = """
        SELECT * FROM {collection}
        WHERE id = $id
        LIMIT 1
    """

    CONTENT_HASH_EXISTS_QUERY: Final[str] = """
        SELECT * FROM {collection}
        WHERE meta_data.content_hash = $content_hash
        LIMIT 1
    """

    DELETE_BY_ID_QUERY: Final[str] = """
        DELETE FROM {collection}
        WHERE id = $id
    """

    DELETE_BY_NAME_QUERY: Final[str] = """
        DELETE FROM {collection}
        WHERE meta_data.name = $name
    """

    DELETE_BY_METADATA_QUERY: Final[str] = """
        DELETE FROM {collection}
        WHERE {conditions}
    """

    DELETE_BY_CONTENT_ID_QUERY: Final[str] = """
        DELETE FROM {collection}
        WHERE content_id = $content_id
    """

    UPSERT_QUERY: Final[str] = """
        UPSERT {thing}
        SET content = $content,
            embedding = $embedding,
            meta_data = $meta_data
    """

    SEARCH_QUERY: Final[str] = """
        SELECT
            content,
            meta_data,
            vector::distance::knn() as distance
        FROM {collection}
        WHERE embedding <|{limit}, {search_ef}|> $query_embedding
        {filter_condition}
        ORDER BY distance ASC
        LIMIT {limit};
    """

    INFO_DB_QUERY: Final[str] = "INFO FOR DB;"
    DROP_TABLE_QUERY: Final[str] = "REMOVE TABLE {collection}"
    DELETE_ALL_QUERY: Final[str] = "DELETE {collection}"

    def __init__(
        self,
        client: Optional[Union[BlockingWsSurrealConnection, BlockingHttpSurrealConnection]] = None,
        async_client: Optional[Union[AsyncWsSurrealConnection, AsyncHttpSurrealConnection]] = None,
        collection: str = "documents",
        distance: Distance = Distance.cosine,
        efc: int = 150,
        m: int = 12,
        search_ef: int = 40,
        embedder: Optional[Embedder] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        id: Optional[str] = None,
    ):
        """Initialize SurrealDB connection.

        Args:
            url: SurrealDB server URL (e.g. ws://localhost:8000/rpc)
            client: A blocking connection, either HTTP or WS
            async_client: An async connection, either HTTP or WS (default: None)
            collection: Collection name to store documents (default: documents)
            distance: Distance metric to use (default: cosine)
            efc: HNSW construction time/accuracy trade-off (default: 150)
            m: HNSW max number of connections per element (default: 12)
            search_ef: HNSW search time/accuracy trade-off (default: 40)
            embedder: Embedder instance for creating embeddings (default: OpenAIEmbedder)

        """
        # Dynamic ID generation based on unique identifiers
        if id is None:
            from agno.utils.string import generate_id

            client_info = str(client) if client else str(async_client) if async_client else "default"
            seed = f"{client_info}#{collection}"
            id = generate_id(seed)

        # Initialize base class with name, description, and generated ID
        super().__init__(id=id, name=name, description=description)

        # Embedder for embedding the document contents
        if embedder is None:
            from agno.knowledge.embedder.openai import OpenAIEmbedder

            embedder = OpenAIEmbedder()
            log_info("Embedder not provided, using OpenAIEmbedder as default.")
        self.embedder: Embedder = embedder
        self.dimensions = self.embedder.dimensions
        self.collection = collection
        # Convert Distance enum to SurrealDB distance type
        self.distance = {Distance.cosine: "COSINE", Distance.l2: "EUCLIDEAN", Distance.max_inner_product: "DOT"}[
            distance
        ]

        self._client: Optional[Union[BlockingHttpSurrealConnection, BlockingWsSurrealConnection]] = client
        self._async_client: Optional[Union[AsyncWsSurrealConnection, AsyncHttpSurrealConnection]] = async_client

        if self._client is None and self._async_client is None:
            msg = "Client and async client are not provided. Please provide one of them."
            raise RuntimeError(msg)

        # HNSW index parameters
        self.efc = efc
        self.m = m
        self.search_ef = search_ef

    @property
    def async_client(self) -> Union[AsyncWsSurrealConnection, AsyncHttpSurrealConnection]:
        """Check if the async client is initialized.

        Raises:
            RuntimeError: If the async client is not initialized.

        Returns:
            The async client.

        """
        if self._async_client is None:
            msg = "Async client is not initialized"
            raise RuntimeError(msg)
        return self._async_client

    @property
    def client(self) -> Union[BlockingHttpSurrealConnection, BlockingWsSurrealConnection]:
        """Check if the client is initialized.

        Returns:
            The client.

        """
        if self._client is None:
            msg = "Client is not initialized"
            raise RuntimeError(msg)
        return self._client

    @staticmethod
    def _build_filter_condition(filters: Optional[Dict[str, Any]] = None) -> str:
        """Build filter condition for queries.

        Args:
            filters: A dictionary of filters to apply to the query.

        Returns:
            A string representing the filter condition.

        """
        if not filters:
            return ""
        conditions = [f"meta_data.{key} = ${key}" for key in filters]
        return "AND " + " AND ".join(conditions)

    # Synchronous methods
    def create(self) -> None:
        """Create the vector collection and index."""
        if not self.exists():
            log_debug(f"Creating collection: {self.collection}")
            query = self.CREATE_TABLE_QUERY.format(
                collection=self.collection,
                distance=self.distance,
                dimensions=self.dimensions,
                efc=self.efc,
                m=self.m,
            )
            self.client.query(query)

    def doc_exists(self, document: Document) -> bool:
        """Check if a document exists by its content.

        Args:
            document: The document to check.

        Returns:
            True if the document exists, False otherwise.

        """
        log_debug(f"Checking if document exists: {document.content}")
        result = self.client.query(
            self.DOC_EXISTS_QUERY.format(collection=self.collection),
            {"content": document.content},
        )
        return bool(self._extract_result(result))

    def name_exists(self, name: str) -> bool:
        """Check if a document exists by its name.

        Args:
            name: The name of the document to check.

        Returns:
            True if the document exists, False otherwise.

        """
        log_debug(f"Checking if document exists: {name}")
        result = self.client.query(self.NAME_EXISTS_QUERY.format(collection=self.collection), {"name": name})
        return bool(self._extract_result(result))

    def id_exists(self, id: str) -> bool:
        """Check if a document exists by its ID.

        Args:
            id: The ID of the document to check.

        Returns:
            True if the document exists, False otherwise.

        """
        log_debug(f"Checking if document exists by ID: {id}")
        result = self.client.query(self.ID_EXISTS_QUERY.format(collection=self.collection), {"id": id})
        return bool(self._extract_result(result))

    def content_hash_exists(self, content_hash: str) -> bool:
        """Check if a document exists by its content hash.

        Args:
            content_hash: The content hash of the document to check.

        Returns:
            True if the document exists, False otherwise.

        """
        log_debug(f"Checking if document exists by content hash: {content_hash}")
        result = self.client.query(
            self.CONTENT_HASH_EXISTS_QUERY.format(collection=self.collection), {"content_hash": content_hash}
        )
        return bool(self._extract_result(result))

    def insert(self, content_hash: str, documents: List[Document], filters: Optional[Dict[str, Any]] = None) -> None:
        """Insert documents into the vector store.

        Args:
            content_hash: The content hash for the documents.
            documents: A list of documents to insert.
            filters: A dictionary of filters to apply to the query.

        """
        for doc in documents:
            doc.embed(embedder=self.embedder)
            meta_data: Dict[str, Any] = doc.meta_data if isinstance(doc.meta_data, dict) else {}
            meta_data["content_hash"] = content_hash
            data: Dict[str, Any] = {"content": doc.content, "embedding": doc.embedding, "meta_data": meta_data}
            if filters:
                data["meta_data"].update(filters)
            self.client.create(self.collection, data)

    def upsert(self, content_hash: str, documents: List[Document], filters: Optional[Dict[str, Any]] = None) -> None:
        """Upsert documents into the vector store.

        Args:
            content_hash: The content hash for the documents.
            documents: A list of documents to upsert.
            filters: A dictionary of filters to apply to the query.

        """
        for doc in documents:
            doc.embed(embedder=self.embedder)
            meta_data: Dict[str, Any] = doc.meta_data if isinstance(doc.meta_data, dict) else {}
            meta_data["content_hash"] = content_hash
            data: Dict[str, Any] = {"content": doc.content, "embedding": doc.embedding, "meta_data": meta_data}
            if filters:
                data["meta_data"].update(filters)
            thing = f"{self.collection}:{doc.id}" if doc.id else self.collection
            self.client.query(self.UPSERT_QUERY.format(thing=thing), data)

    def search(self, query: str, limit: int = 5, filters: Optional[Dict[str, Any]] = None) -> List[Document]:
        """Search for similar documents.

        Args:
            query: The query to search for.
            limit: The maximum number of documents to return.
            filters: A dictionary of filters to apply to the query.

        Returns:
            A list of documents that are similar to the query.

        """
        query_embedding = self.embedder.get_embedding(query)
        if query_embedding is None:
            log_error(f"Error getting embedding for Query: {query}")
            return []

        filter_condition = self._build_filter_condition(filters)
        log_debug(f"Filter condition: {filter_condition}")
        search_query = self.SEARCH_QUERY.format(
            collection=self.collection,
            limit=limit,
            search_ef=self.search_ef,
            filter_condition=filter_condition,
            distance=self.distance,
        )
        log_debug(f"Search query: {search_query}")
        response = self.client.query(
            search_query,
            {"query_embedding": query_embedding, **filters} if filters else {"query_embedding": query_embedding},
        )
        log_debug(f"Search response: {response}")

        documents = []
        for item in response:
            if isinstance(item, dict):
                doc = Document(
                    content=item.get("content", ""),
                    embedding=item.get("embedding", []),
                    meta_data=item.get("meta_data", {}),
                    embedder=self.embedder,
                )
                documents.append(doc)
        log_debug(f"Found {len(documents)} documents")
        return documents

    def drop(self) -> None:
        """Drop the vector collection."""
        log_debug(f"Dropping collection: {self.collection}")
        self.client.query(self.DROP_TABLE_QUERY.format(collection=self.collection))

    def exists(self) -> bool:
        """Check if the vector collection exists.

        Returns:
            True if the collection exists, False otherwise.

        """
        log_debug(f"Checking if collection exists: {self.collection}")
        response = self.client.query(self.INFO_DB_QUERY)
        result = self._extract_result(response)
        if isinstance(result, dict) and "tables" in result:
            return self.collection in result["tables"]
        return False

    def delete(self) -> bool:
        """Delete all documents from the vector store.

        Returns:
            True if the collection was deleted, False otherwise.

        """
        self.client.query(self.DELETE_ALL_QUERY.format(collection=self.collection))
        return True

    def delete_by_id(self, id: str) -> bool:
        """Delete a document by its ID.

        Args:
            id: The ID of the document to delete.

        Returns:
            True if the document was deleted, False otherwise.

        """
        log_debug(f"Deleting document by ID: {id}")
        result = self.client.query(self.DELETE_BY_ID_QUERY.format(collection=self.collection), {"id": id})
        return bool(result)

    def delete_by_name(self, name: str) -> bool:
        """Delete documents by their name.

        Args:
            name: The name of the documents to delete.

        Returns:
            True if documents were deleted, False otherwise.

        """
        log_debug(f"Deleting documents by name: {name}")
        result = self.client.query(self.DELETE_BY_NAME_QUERY.format(collection=self.collection), {"name": name})
        return bool(result)

    def delete_by_metadata(self, metadata: Dict[str, Any]) -> bool:
        """Delete documents by their metadata.

        Args:
            metadata: The metadata to match for deletion.

        Returns:
            True if documents were deleted, False otherwise.

        """
        log_debug(f"Deleting documents by metadata: {metadata}")
        conditions = [f"meta_data.{key} = ${key}" for key in metadata.keys()]
        conditions_str = " AND ".join(conditions)
        query = self.DELETE_BY_METADATA_QUERY.format(collection=self.collection, conditions=conditions_str)
        result = self.client.query(query, metadata)
        return bool(result)

    def delete_by_content_id(self, content_id: str) -> bool:
        """Delete documents by their content ID.

        Args:
            content_id: The content ID of the documents to delete.

        Returns:
            True if documents were deleted, False otherwise.

        """
        log_debug(f"Deleting documents by content ID: {content_id}")
        result = self.client.query(
            self.DELETE_BY_CONTENT_ID_QUERY.format(collection=self.collection), {"content_id": content_id}
        )
        return bool(result)

    @staticmethod
    def _extract_result(query_result: Union[List[Dict[str, Any]], Dict[str, Any]]) -> Union[List[Any], Dict[str, Any]]:
        """Extract the actual result from SurrealDB query response.

        Args:
            query_result: The query result from SurrealDB.

        Returns:
            The actual result from SurrealDB query response.

        """
        log_debug(f"Query result: {query_result}")
        if isinstance(query_result, dict):
            return query_result
        if isinstance(query_result, list):
            if len(query_result) > 0:
                return query_result[0].get("result", {})
            return []
        return []

    async def async_create(self) -> None:
        """Create the vector collection and index asynchronously."""
        log_debug(f"Creating collection: {self.collection}")
        await self.async_client.query(
            self.CREATE_TABLE_QUERY.format(
                collection=self.collection,
                distance=self.distance,
                dimensions=self.dimensions,
                efc=self.efc,
                m=self.m,
            ),
        )

    async def async_doc_exists(self, document: Document) -> bool:
        """Check if a document exists by its content asynchronously.

        Returns:
            True if the document exists, False otherwise.

        """
        response = await self.async_client.query(
            self.DOC_EXISTS_QUERY.format(collection=self.collection),
            {"content": document.content},
        )
        return bool(self._extract_result(response))

    async def async_name_exists(self, name: str) -> bool:
        """Check if a document exists by its name asynchronously.

        Returns:
            True if the document exists, False otherwise.

        """
        response = await self.async_client.query(
            self.NAME_EXISTS_QUERY.format(collection=self.collection),
            {"name": name},
        )
        return bool(self._extract_result(response))

    async def async_insert(
        self, content_hash: str, documents: List[Document], filters: Optional[Dict[str, Any]] = None
    ) -> None:
        """Insert documents into the vector store asynchronously.

        Args:
            content_hash: The content hash for the documents.
            documents: A list of documents to insert.
            filters: A dictionary of filters to apply to the query.

        """
        for doc in documents:
            doc.embed(embedder=self.embedder)
            meta_data: Dict[str, Any] = doc.meta_data if isinstance(doc.meta_data, dict) else {}
            meta_data["content_hash"] = content_hash
            data: Dict[str, Any] = {"content": doc.content, "embedding": doc.embedding, "meta_data": meta_data}
            if filters:
                data["meta_data"].update(filters)
            log_debug(f"Inserting document asynchronously: {doc.name} ({doc.meta_data})")
            await self.async_client.create(self.collection, data)

    async def async_upsert(
        self, content_hash: str, documents: List[Document], filters: Optional[Dict[str, Any]] = None
    ) -> None:
        """Upsert documents into the vector store asynchronously.

        Args:
            content_hash: The content hash for the documents.
            documents: A list of documents to upsert.
            filters: A dictionary of filters to apply to the query.

        """
        for doc in documents:
            doc.embed(embedder=self.embedder)
            meta_data: Dict[str, Any] = doc.meta_data if isinstance(doc.meta_data, dict) else {}
            meta_data["content_hash"] = content_hash
            data: Dict[str, Any] = {"content": doc.content, "embedding": doc.embedding, "meta_data": meta_data}
            if filters:
                data["meta_data"].update(filters)
            log_debug(f"Upserting document asynchronously: {doc.name} ({doc.meta_data})")
            thing = f"{self.collection}:{doc.id}" if doc.id else self.collection
            await self.async_client.query(self.UPSERT_QUERY.format(thing=thing), data)

    async def async_search(
        self,
        query: str,
        limit: int = 5,
        filters: Optional[Dict[str, Any]] = None,
    ) -> List[Document]:
        """Search for similar documents asynchronously.

        Args:
            query: The query to search for.
            limit: The maximum number of documents to return.
            filters: A dictionary of filters to apply to the query.

        Returns:
            A list of documents that are similar to the query.

        """
        query_embedding = self.embedder.get_embedding(query)
        if query_embedding is None:
            log_error(f"Error getting embedding for Query: {query}")
            return []

        filter_condition = self._build_filter_condition(filters)
        search_query = self.SEARCH_QUERY.format(
            collection=self.collection,
            limit=limit,
            search_ef=self.search_ef,
            filter_condition=filter_condition,
            distance=self.distance,
        )
        response = await self.async_client.query(
            search_query,
            {"query_embedding": query_embedding, **filters} if filters else {"query_embedding": query_embedding},
        )
        log_debug(f"Search response: {response}")
        documents = []
        for item in response:
            if isinstance(item, dict):
                doc = Document(
                    content=item.get("content", ""),
                    embedding=item.get("embedding", []),
                    meta_data=item.get("meta_data", {}),
                    embedder=self.embedder,
                )
                documents.append(doc)
        log_debug(f"Found {len(documents)} documents asynchronously")
        return documents

    async def async_drop(self) -> None:
        """Drop the vector collection asynchronously."""
        log_debug(f"Dropping collection: {self.collection}")
        await self.async_client.query(self.DROP_TABLE_QUERY.format(collection=self.collection))

    async def async_exists(self) -> bool:
        """Check if the vector collection exists asynchronously.

        Returns:
            True if the collection exists, False otherwise.

        """
        log_debug(f"Checking if collection exists: {self.collection}")
        response = await self.async_client.query(self.INFO_DB_QUERY)
        result = self._extract_result(response)
        if isinstance(result, dict) and "tables" in result:
            return self.collection in result["tables"]
        return False

    @staticmethod
    def upsert_available() -> bool:
        """Check if upsert is available.

        Returns:
            True if upsert is available, False otherwise.

        """
        return True

    def update_metadata(self, content_id: str, metadata: Dict[str, Any]) -> None:
        """
        Update the metadata for documents with the given content_id.

        Args:
            content_id (str): The content ID to update
            metadata (Dict[str, Any]): The metadata to update
        """
        try:
            # Query for documents with the given content_id
            query = f"SELECT * FROM {self.collection} WHERE content_id = $content_id"
            result = self.client.query(query, {"content_id": content_id})

            if not result or not result[0].get("result"):
                log_debug(f"No documents found with content_id: {content_id}")
                return

            documents = result[0]["result"]
            updated_count = 0

            # Update each matching document
            for doc in documents:
                doc_id = doc["id"]
                current_metadata = doc.get("meta_data", {})
                current_filters = doc.get("filters", {})

                # Merge existing metadata with new metadata
                if isinstance(current_metadata, dict):
                    updated_metadata = current_metadata.copy()
                    updated_metadata.update(metadata)
                else:
                    updated_metadata = metadata

                # Merge existing filters with new metadata
                if isinstance(current_filters, dict):
                    updated_filters = current_filters.copy()
                    updated_filters.update(metadata)
                else:
                    updated_filters = metadata

                # Update the document
                update_query = f"UPDATE {doc_id} SET meta_data = $metadata, filters = $filters"
                self.client.query(update_query, {"metadata": updated_metadata, "filters": updated_filters})
                updated_count += 1

            log_debug(f"Updated metadata for {updated_count} documents with content_id: {content_id}")

        except Exception as e:
            log_error(f"Error updating metadata for content_id '{content_id}': {e}")
            raise

    def get_supported_search_types(self) -> List[str]:
        """Get the supported search types for this vector database."""
        return []  # SurrealDb doesn't use SearchType enum



import json
from os import getenv
from typing import Any, Dict, List, Optional

from agno.tools import Toolkit
from agno.utils.log import log_error

try:
    from firecrawl import FirecrawlApp  # type: ignore[attr-defined]
    from firecrawl.types import ScrapeOptions
except ImportError:
    raise ImportError("`firecrawl-py` not installed. Please install using `pip install firecrawl-py`")


class CustomJSONEncoder(json.JSONEncoder):
    """Custom JSON encoder that handles non-serializable types by converting them to strings."""

    def default(self, obj):
        try:
            return super().default(obj)
        except TypeError:
            return str(obj)


class FirecrawlTools(Toolkit):
    """
    Firecrawl is a tool for scraping and crawling websites.

    Args:
        api_key (Optional[str]): The API key to use for the Firecrawl app.
        enable_scrape (bool): Enable website scraping functionality. Default is True.
        enable_crawl (bool): Enable website crawling functionality. Default is False.
        enable_mapping (bool): Enable website mapping functionality. Default is False.
        enable_search (bool): Enable web search functionality. Default is False.
        all (bool): Enable all tools. Overrides individual flags when True. Default is False.
        formats (Optional[List[str]]): The formats to use for the Firecrawl app.
        limit (int): The maximum number of pages to crawl.
        poll_interval (int): Polling interval for crawl operations.
        search_params (Optional[Dict[str, Any]]): Parameters for search operations.
        api_url (Optional[str]): The API URL to use for the Firecrawl app.
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        enable_scrape: bool = True,
        enable_crawl: bool = False,
        enable_mapping: bool = False,
        enable_search: bool = False,
        all: bool = False,
        formats: Optional[List[str]] = None,
        limit: int = 10,
        poll_interval: int = 30,
        search_params: Optional[Dict[str, Any]] = None,
        api_url: Optional[str] = "https://api.firecrawl.dev",
        **kwargs,
    ):
        self.api_key: Optional[str] = api_key or getenv("FIRECRAWL_API_KEY")
        if not self.api_key:
            log_error("FIRECRAWL_API_KEY not set. Please set the FIRECRAWL_API_KEY environment variable.")

        self.formats: Optional[List[str]] = formats
        self.limit: int = limit
        self.poll_interval: int = poll_interval
        self.app: FirecrawlApp = FirecrawlApp(api_key=self.api_key, api_url=api_url)
        self.search_params = search_params

        tools: List[Any] = []
        if all or enable_scrape:
            tools.append(self.scrape_website)
        if all or enable_crawl:
            tools.append(self.crawl_website)
        if all or enable_mapping:
            tools.append(self.map_website)
        if all or enable_search:
            tools.append(self.search_web)

        super().__init__(name="firecrawl_tools", tools=tools, **kwargs)

    def scrape_website(self, url: str) -> str:
        """Use this function to scrape a website using Firecrawl.

        Args:
            url (str): The URL to scrape.
        """
        params = {}
        if self.formats:
            params["formats"] = self.formats

        scrape_result = self.app.scrape(url, **params)
        return json.dumps(scrape_result.model_dump(), cls=CustomJSONEncoder)

    def crawl_website(self, url: str, limit: Optional[int] = None) -> str:
        """Use this function to Crawls a website using Firecrawl.

        Args:
            url (str): The URL to crawl.
            limit (int): The maximum number of pages to crawl

        Returns:
            The results of the crawling.
        """
        params: Dict[str, Any] = {}
        if self.limit or limit:
            params["limit"] = self.limit or limit
        if self.formats:
            params["scrape_options"] = ScrapeOptions(formats=self.formats)  # type: ignore

        params["poll_interval"] = self.poll_interval

        crawl_result = self.app.crawl(url, **params)
        return json.dumps(crawl_result.model_dump(), cls=CustomJSONEncoder)

    def map_website(self, url: str) -> str:
        """Use this function to Map a website using Firecrawl.

        Args:
            url (str): The URL to map.

        """
        map_result = self.app.map(url)
        return json.dumps(map_result.model_dump(), cls=CustomJSONEncoder)

    def search_web(self, query: str, limit: Optional[int] = None):
        """Use this function to search for the web using Firecrawl.

        Args:
            query (str): The query to search for.
            limit (int): The maximum number of results to return.
        """
        params: Dict[str, Any] = {}
        if self.limit or limit:
            params["limit"] = self.limit or limit
        if self.formats:
            params["scrape_options"] = ScrapeOptions(formats=self.formats)  # type: ignore
        if self.search_params:
            params.update(self.search_params)

        search_result = self.app.search(query, **params)
        if search_result.success:
            return json.dumps(search_result.data, cls=CustomJSONEncoder)
        else:
            return "Error searching with the Firecrawl tool: " + search_result.error



import os
from typing import Optional, Union

import httpx
import requests

from agno.utils.log import log_debug, log_error


def get_access_token() -> str:
    access_token = os.getenv("WHATSAPP_ACCESS_TOKEN")
    if not access_token:
        raise ValueError("WHATSAPP_ACCESS_TOKEN is not set")
    return access_token


def get_phone_number_id() -> str:
    phone_number_id = os.getenv("WHATSAPP_PHONE_NUMBER_ID")
    if not phone_number_id:
        raise ValueError("WHATSAPP_PHONE_NUMBER_ID is not set")
    return phone_number_id


def get_media(media_id: str) -> Union[dict, bytes]:
    """
    Sends a GET request to the Facebook Graph API to retrieve media information.

    Args:
        media_id (str): The ID of the media to retrieve.
    """
    url = f"https://graph.facebook.com/v22.0/{media_id}"

    access_token = get_access_token()

    headers = {"Authorization": f"Bearer {access_token}"}
    try:
        response = requests.get(url, headers=headers)
        response.raise_for_status()  # Raise an HTTPError for bad responses (4xx and 5xx)
        data = response.json()

        media_url = data.get("url")
    except requests.exceptions.RequestException as e:
        return {"error": str(e)}

    try:
        response = requests.get(media_url, headers=headers)
        response.raise_for_status()  # Raise an HTTPError for bad responses (4xx and 5xx)
        data = response.content
        return data
    except requests.exceptions.RequestException as e:
        return {"error": str(e)}


async def get_media_async(media_id: str) -> Union[dict, bytes]:
    """
    Sends a GET request to the Facebook Graph API to retrieve media information.

    Args:
        media_id (str): The ID of the media to retrieve.
    """
    url = f"https://graph.facebook.com/v22.0/{media_id}"

    access_token = get_access_token()

    headers = {"Authorization": f"Bearer {access_token}"}
    try:
        async with httpx.AsyncClient() as client:
            response = await client.get(url, headers=headers)
            response.raise_for_status()  # Raise an HTTPError for bad responses (4xx and 5xx)
            data = response.json()

        media_url = data.get("url")
    except httpx.HTTPStatusError as e:
        return {"error": str(e)}

    try:
        async with httpx.AsyncClient() as client:
            response = await client.get(media_url, headers=headers)
            response.raise_for_status()  # Raise an HTTPError for bad responses (4xx and 5xx)
            data = response.content
        return data
    except httpx.HTTPStatusError as e:
        return {"error": str(e)}


def upload_media(media_data: bytes, mime_type: str, filename: str = "file"):
    """
    Sends a POST request to the Facebook Graph API to upload media for WhatsApp.

    Args:
        media_data: Bytes buffer containing the file data
        mime_type (str): The MIME type of the file
        filename (str): The name to use for the file in the upload. Defaults to "file"
    """
    phone_number_id = get_phone_number_id()

    url = f"https://graph.facebook.com/v22.0/{phone_number_id}/media"

    access_token = get_access_token()

    headers = {"Authorization": f"Bearer {access_token}"}

    data = {"messaging_product": "whatsapp", "type": mime_type}
    try:
        from io import BytesIO

        file_data = BytesIO(media_data)
        files = {"file": (filename, file_data, mime_type)}

        response = requests.post(url, headers=headers, data=data, files=files)
        response.raise_for_status()  # Raise an error for bad responses
        json_resp = response.json()
        media_id = json_resp.get("id")
        if not media_id:
            return {"error": "Media ID not found in response", "response": json_resp}
        return media_id
    except requests.exceptions.RequestException as e:
        return {"error": str(e)}
    except Exception as e:
        return {"error": str(e)}


async def upload_media_async(media_data: bytes, mime_type: str, filename: str = "file"):
    """
    Sends a POST request to the Facebook Graph API to upload media for WhatsApp.

    Args:
        media_data: Bytes buffer containing the file data
        mime_type (str): The MIME type of the file
        filename (str): The name to use for the file in the upload. Defaults to "file"
    """
    phone_number_id = get_phone_number_id()

    url = f"https://graph.facebook.com/v22.0/{phone_number_id}/media"

    access_token = get_access_token()

    headers = {"Authorization": f"Bearer {access_token}"}

    data = {"messaging_product": "whatsapp", "type": mime_type}
    try:
        from io import BytesIO

        file_data = BytesIO(media_data)
        files = {"file": (filename, file_data, mime_type)}

        async with httpx.AsyncClient() as client:
            response = await client.post(url, headers=headers, data=data, files=files)
            response.raise_for_status()  # Raise an error for bad responses
            json_resp = response.json()
            media_id = json_resp.get("id")
            if not media_id:
                return {"error": "Media ID not found in response", "response": json_resp}
            return media_id
    except httpx.HTTPStatusError as e:
        return {"error": str(e)}
    except Exception as e:
        return {"error": str(e)}


async def send_image_message_async(
    media_id: str,
    recipient: str,
    text: Optional[str] = None,
):
    """Send an image message to a WhatsApp user (asynchronous version).

    Args:
        media_id: The media id for the image to send
        recipient: Recipient's WhatsApp ID or phone number (e.g., "+1234567890").
        text: Caption for the image

    Returns:
        Success message with message ID
    """
    log_debug(f"Sending WhatsApp image to {recipient}: {text}")
    phone_number_id = get_phone_number_id()

    url = f"https://graph.facebook.com/v22.0/{phone_number_id}/messages"

    access_token = get_access_token()

    headers = {"Authorization": f"Bearer {access_token}"}

    data = {
        "messaging_product": "whatsapp",
        "recipient_type": "individual",
        "to": recipient,
        "type": "image",
        "image": {"id": media_id, "caption": text},
    }

    try:
        async with httpx.AsyncClient() as client:
            import json

            log_debug(f"Request data: {json.dumps(data, indent=2)}")
            response = await client.post(url, headers=headers, json=data)
            response.raise_for_status()
            log_debug(f"Response: {response.text}")

    except httpx.HTTPStatusError as e:
        log_error(f"Failed to send WhatsApp image message: {e}")
        log_error(f"Error response: {e.response.text if hasattr(e, 'response') else 'No response text'}")
        raise
    except Exception as e:
        log_error(f"Unexpected error sending WhatsApp image message: {str(e)}")
        raise


def send_image_message(
    media_id: str,
    recipient: str,
    text: Optional[str] = None,
):
    """Send an image message to a WhatsApp user (synchronous version).

    Args:
        image: The media id for the image to send
        recipient: Recipient's WhatsApp ID or phone number (e.g., "+1234567890").
        text: Caption for the image

    Returns:
        Success message with message ID
    """
    log_debug(f"Sending WhatsApp image to {recipient}: {text}")
    phone_number_id = get_phone_number_id()

    url = f"https://graph.facebook.com/v22.0/{phone_number_id}/messages"

    access_token = get_access_token()

    headers = {"Authorization": f"Bearer {access_token}"}

    data = {
        "messaging_product": "whatsapp",
        "recipient_type": "individual",
        "to": recipient,
        "type": "image",
        "image": {"id": media_id, "caption": text},
    }

    try:
        import json

        log_debug(f"Request data: {json.dumps(data, indent=2)}")
        response = requests.post(url, headers=headers, json=data)
        response.raise_for_status()
        log_debug(f"Response: {response.text}")
    except requests.exceptions.RequestException as e:
        log_error(f"Failed to send WhatsApp image message: {e}")
        log_error(f"Error response: {e.response.text if hasattr(e, 'response') else 'No response text'}")  # type: ignore
        raise
    except Exception as e:
        log_error(f"Unexpected error sending WhatsApp image message: {str(e)}")
        raise


def typing_indicator(message_id: Optional[str] = None):
    if not message_id:
        return

    phone_number_id = get_phone_number_id()

    url = f"https://graph.facebook.com/v22.0/{phone_number_id}/messages"

    access_token = get_access_token()

    headers = {"Authorization": f"Bearer {access_token}"}
    data = {
        "messaging_product": "whatsapp",
        "status": "read",
        "message_id": f"{message_id}",
        "typing_indicator": {"type": "text"},
    }
    try:
        response = requests.post(url, headers=headers, data=data)
        response.raise_for_status()  # Raise an HTTPError for bad responses (4xx and 5xx)
    except requests.exceptions.RequestException as e:
        return {"error": str(e)}


async def typing_indicator_async(message_id: Optional[str] = None):
    if not message_id:
        return

    phone_number_id = get_phone_number_id()

    url = f"https://graph.facebook.com/v22.0/{phone_number_id}/messages"

    access_token = get_access_token()

    headers = {"Authorization": f"Bearer {access_token}"}
    data = {
        "messaging_product": "whatsapp",
        "status": "read",
        "message_id": f"{message_id}",
        "typing_indicator": {"type": "text"},
    }
    try:
        async with httpx.AsyncClient() as client:
            response = await client.post(url, headers=headers, data=data)
            response.raise_for_status()  # Raise an HTTPError for bad responses (4xx and 5xx)
    except httpx.HTTPStatusError as e:
        return {"error": str(e)}



from datetime import datetime, timezone
from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field


class ContentStatus(str, Enum):
    """Enumeration of possible content processing statuses."""

    PROCESSING = "processing"
    COMPLETED = "completed"
    FAILED = "failed"


class ContentStatusResponse(BaseModel):
    """Response model for content status endpoint."""

    status: ContentStatus
    status_message: str = ""


class ContentResponseSchema(BaseModel):
    id: str
    name: Optional[str] = None
    description: Optional[str] = None
    type: Optional[str] = None
    size: Optional[str] = None
    linked_to: Optional[str] = None
    metadata: Optional[dict] = None
    access_count: Optional[int] = None
    status: Optional[ContentStatus] = None
    status_message: Optional[str] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None

    @classmethod
    def from_dict(cls, content: Dict[str, Any]) -> "ContentResponseSchema":
        status = content.get("status")
        if isinstance(status, str):
            try:
                status = ContentStatus(status.lower())
            except ValueError:
                # Handle legacy or unknown statuses gracefully
                if "failed" in status.lower():
                    status = ContentStatus.FAILED
                elif "completed" in status.lower():
                    status = ContentStatus.COMPLETED
                else:
                    status = ContentStatus.PROCESSING
        elif status is None:
            status = ContentStatus.PROCESSING  # Default for None values

        # Helper function to safely parse timestamps
        def parse_timestamp(timestamp_value):
            if timestamp_value is None:
                return None
            try:
                # If it's already a datetime object, return it
                if isinstance(timestamp_value, datetime):
                    return timestamp_value
                # If it's a string, try to parse it as ISO format first
                if isinstance(timestamp_value, str):
                    try:
                        return datetime.fromisoformat(timestamp_value.replace("Z", "+00:00"))
                    except ValueError:
                        # Try to parse as float/int timestamp
                        timestamp_value = float(timestamp_value)
                # If it's a number, use fromtimestamp
                return datetime.fromtimestamp(timestamp_value, tz=timezone.utc)
            except (ValueError, TypeError, OSError):
                # If all parsing fails, return None
                return None

        return cls(
            id=content.get("id"),  # type: ignore
            name=content.get("name"),
            description=content.get("description"),
            type=content.get("file_type"),
            size=str(content.get("size")) if content.get("size") else "0",
            metadata=content.get("metadata"),
            status=status,
            status_message=content.get("status_message"),
            created_at=parse_timestamp(content.get("created_at")),
            updated_at=parse_timestamp(content.get("updated_at")),
            # TODO: These fields are not available in the Content class. Fix the inconsistency
            access_count=None,
            linked_to=None,
        )


class ContentUpdateSchema(BaseModel):
    """Schema for updating content."""

    name: Optional[str] = Field(None, description="Content name", min_length=1, max_length=255)
    description: Optional[str] = Field(None, description="Content description", max_length=1000)
    metadata: Optional[Dict[str, Any]] = Field(None, description="Content metadata as key-value pairs")
    reader_id: Optional[str] = Field(None, description="ID of the reader to use for processing", min_length=1)


class ReaderSchema(BaseModel):
    id: str
    name: Optional[str] = None
    description: Optional[str] = None
    chunkers: Optional[List[str]] = None


class ChunkerSchema(BaseModel):
    key: str
    name: Optional[str] = None
    description: Optional[str] = None


class VectorDbSchema(BaseModel):
    id: str
    name: Optional[str] = None
    description: Optional[str] = None
    search_types: Optional[List[str]] = None


class VectorSearchResult(BaseModel):
    """Schema for search result documents."""

    id: str
    content: str
    name: Optional[str] = None
    meta_data: Optional[Dict[str, Any]] = None
    usage: Optional[Dict[str, Any]] = None
    reranking_score: Optional[float] = None
    content_id: Optional[str] = None
    content_origin: Optional[str] = None
    size: Optional[int] = None

    @classmethod
    def from_document(cls, document) -> "VectorSearchResult":
        """Convert a Document object to a serializable VectorSearchResult."""
        return cls(
            id=document.id,
            content=document.content,
            name=getattr(document, "name", None),
            meta_data=getattr(document, "meta_data", None),
            usage=getattr(document, "usage", None),
            reranking_score=getattr(document, "reranking_score", None),
            content_id=getattr(document, "content_id", None),
            content_origin=getattr(document, "content_origin", None),
            size=getattr(document, "size", None),
        )


class VectorSearchRequestSchema(BaseModel):
    """Schema for vector search request."""

    class Meta(BaseModel):
        """Inline metadata schema for pagination."""

        limit: Optional[int] = Field(20, description="Number of results per page", ge=1, le=100)
        page: Optional[int] = Field(1, description="Page number", ge=1)

    query: str = Field(..., description="The search query")
    db_id: Optional[str] = Field(None, description="The content database id")
    vector_db_ids: Optional[List[str]] = Field(None, description="List of vector database ids to search in")
    search_type: Optional[str] = Field(None, description="The type of search to perform")
    max_results: Optional[int] = Field(None, description="The maximum number of results to return")
    filters: Optional[Dict[str, Any]] = Field(None, description="The filters to apply to the search")
    meta: Optional[Meta] = Field(
        None, description="Pagination metadata. Limit and page number to return a subset of results."
    )


class ConfigResponseSchema(BaseModel):
    readers: Optional[Dict[str, ReaderSchema]] = None
    readersForType: Optional[Dict[str, List[str]]] = None
    chunkers: Optional[Dict[str, ChunkerSchema]] = None
    filters: Optional[List[str]] = None
    vector_dbs: Optional[List[VectorDbSchema]] = None



import inspect
from typing import Any, Dict, List, Optional

from agno.knowledge.chunking.strategy import ChunkingStrategy
from agno.knowledge.document.base import Document
from agno.knowledge.embedder.base import Embedder
from agno.knowledge.embedder.openai import OpenAIEmbedder


class SemanticChunking(ChunkingStrategy):
    """Chunking strategy that splits text into semantic chunks using chonkie"""

    def __init__(self, embedder: Optional[Embedder] = None, chunk_size: int = 5000, similarity_threshold: float = 0.5):
        self.embedder = embedder or OpenAIEmbedder(id="text-embedding-3-small")  # type: ignore
        self.chunk_size = chunk_size
        self.similarity_threshold = similarity_threshold
        self.chunker = None  # Will be initialized lazily when needed

    def _initialize_chunker(self):
        """Lazily initialize the chunker with chonkie dependency."""
        if self.chunker is None:
            try:
                from chonkie import SemanticChunker
            except ImportError:
                raise ImportError(
                    "`chonkie` is required for semantic chunking. "
                    "Please install it using `pip install chonkie` to use SemanticChunking."
                )

            # Build arguments dynamically based on chonkie's supported signature
            params: Dict[str, Any] = {
                "chunk_size": self.chunk_size,
                "threshold": self.similarity_threshold,
            }

            try:
                sig = inspect.signature(SemanticChunker)
                param_names = set(sig.parameters.keys())

                # Prefer passing a callable to avoid Chonkie initializing its own client
                if "embedding_fn" in param_names:
                    params["embedding_fn"] = self.embedder.get_embedding  # type: ignore[attr-defined]
                    # If chonkie allows specifying dimensions, provide them
                    if "embedding_dimensions" in param_names and getattr(self.embedder, "dimensions", None):
                        params["embedding_dimensions"] = self.embedder.dimensions  # type: ignore[attr-defined]
                elif "embedder" in param_names:
                    # Some versions may accept an embedder object directly
                    params["embedder"] = self.embedder
                else:
                    # Fallback to model id
                    params["embedding_model"] = getattr(self.embedder, "id", None) or "text-embedding-3-small"

                self.chunker = SemanticChunker(**params)
            except Exception:
                # As a final fallback, use the original behavior
                self.chunker = SemanticChunker(
                    embedding_model=getattr(self.embedder, "id", None) or "text-embedding-3-small",
                    chunk_size=self.chunk_size,
                    threshold=self.similarity_threshold,
                )

    def chunk(self, document: Document) -> List[Document]:
        """Split document into semantic chunks using chonkie"""
        if not document.content:
            return [document]

        # Ensure chunker is initialized (will raise ImportError if chonkie is missing)
        self._initialize_chunker()

        # Use chonkie to split into semantic chunks
        if self.chunker is None:
            raise RuntimeError("Chunker failed to initialize")

        chunks = self.chunker.chunk(self.clean_text(document.content))

        # Convert chunks to Documents
        chunked_documents: List[Document] = []
        for i, chunk in enumerate(chunks, 1):
            meta_data = document.meta_data.copy()
            meta_data["chunk"] = i
            chunk_id = f"{document.id}_{i}" if document.id else None
            meta_data["chunk_size"] = len(chunk.text)

            chunked_documents.append(Document(id=chunk_id, name=document.name, meta_data=meta_data, content=chunk.text))

        return chunked_documents



"""Async router handling exposing an Agno Agent or Team in an AG-UI compatible format."""

import logging
import uuid
from typing import AsyncIterator, Optional

from ag_ui.core import (
    BaseEvent,
    EventType,
    RunAgentInput,
    RunErrorEvent,
    RunStartedEvent,
)
from ag_ui.encoder import EventEncoder
from fastapi import APIRouter
from fastapi.responses import StreamingResponse

from agno.agent.agent import Agent
from agno.os.interfaces.agui.utils import (
    async_stream_agno_response_as_agui_events,
    convert_agui_messages_to_agno_messages,
)
from agno.team.team import Team

logger = logging.getLogger(__name__)


async def run_agent(agent: Agent, run_input: RunAgentInput) -> AsyncIterator[BaseEvent]:
    """Run the contextual Agent, mapping AG-UI input messages to Agno format, and streaming the response in AG-UI format."""
    run_id = run_input.run_id or str(uuid.uuid4())

    try:
        # Preparing the input for the Agent and emitting the run started event
        messages = convert_agui_messages_to_agno_messages(run_input.messages or [])
        yield RunStartedEvent(type=EventType.RUN_STARTED, thread_id=run_input.thread_id, run_id=run_id)

        # Look for user_id in run_input.forwarded_props
        user_id = None
        if run_input.forwarded_props and isinstance(run_input.forwarded_props, dict):
            user_id = run_input.forwarded_props.get("user_id")

        # Request streaming response from agent
        response_stream = agent.arun(
            input=messages,
            session_id=run_input.thread_id,
            stream=True,
            stream_intermediate_steps=True,
            user_id=user_id,
        )

        # Stream the response content in AG-UI format
        async for event in async_stream_agno_response_as_agui_events(
            response_stream=response_stream,  # type: ignore
            thread_id=run_input.thread_id,
            run_id=run_id,
        ):
            yield event

    # Emit a RunErrorEvent if any error occurs
    except Exception as e:
        logger.error(f"Error running agent: {e}", exc_info=True)
        yield RunErrorEvent(type=EventType.RUN_ERROR, message=str(e))


async def run_team(team: Team, input: RunAgentInput) -> AsyncIterator[BaseEvent]:
    """Run the contextual Team, mapping AG-UI input messages to Agno format, and streaming the response in AG-UI format."""
    run_id = input.run_id or str(uuid.uuid4())
    try:
        # Extract the last user message for team execution
        messages = convert_agui_messages_to_agno_messages(input.messages or [])
        yield RunStartedEvent(type=EventType.RUN_STARTED, thread_id=input.thread_id, run_id=run_id)

        # Look for user_id in input.forwarded_props
        user_id = None
        if input.forwarded_props and isinstance(input.forwarded_props, dict):
            user_id = input.forwarded_props.get("user_id")

        # Request streaming response from team
        response_stream = team.arun(
            input=messages,
            session_id=input.thread_id,
            stream=True,
            stream_intermediate_steps=True,
            user_id=user_id,
        )

        # Stream the response content in AG-UI format
        async for event in async_stream_agno_response_as_agui_events(
            response_stream=response_stream, thread_id=input.thread_id, run_id=run_id
        ):
            yield event

    except Exception as e:
        logger.error(f"Error running team: {e}", exc_info=True)
        yield RunErrorEvent(type=EventType.RUN_ERROR, message=str(e))


def attach_routes(router: APIRouter, agent: Optional[Agent] = None, team: Optional[Team] = None) -> APIRouter:
    if agent is None and team is None:
        raise ValueError("Either agent or team must be provided.")

    encoder = EventEncoder()

    @router.post(
        "/agui",
        name="run_agent",
    )
    async def run_agent_agui(run_input: RunAgentInput):
        async def event_generator():
            if agent:
                async for event in run_agent(agent, run_input):
                    encoded_event = encoder.encode(event)
                    yield encoded_event
            elif team:
                async for event in run_team(team, run_input):
                    encoded_event = encoder.encode(event)
                    yield encoded_event

        return StreamingResponse(
            event_generator(),
            media_type="text/event-stream",
            headers={
                "Cache-Control": "no-cache",
                "Connection": "keep-alive",
                "Access-Control-Allow-Origin": "*",
                "Access-Control-Allow-Methods": "POST, GET, OPTIONS",
                "Access-Control-Allow-Headers": "*",
            },
        )

    @router.get("/status")
    async def get_status():
        return {"status": "available"}

    return router



import json
from typing import Any, List, Optional

from agno.knowledge.document import Document
from agno.knowledge.knowledge import Knowledge
from agno.tools import Toolkit
from agno.utils.log import log_debug


class WebsiteTools(Toolkit):
    def __init__(
        self,
        knowledge: Optional[Knowledge] = None,
        **kwargs,
    ):
        self.knowledge: Optional[Knowledge] = knowledge

        tools: List[Any] = []
        if self.knowledge is not None:
            tools.append(self.add_website_to_knowledge)
        else:
            tools.append(self.read_url)

        super().__init__(name="website_tools", tools=tools, **kwargs)

    def add_website_to_knowledge(self, url: str) -> str:
        """This function adds a websites content to the knowledge base.
        NOTE: The website must start with https:// and should be a valid website.

        USE THIS FUNCTION TO GET INFORMATION ABOUT PRODUCTS FROM THE INTERNET.

        :param url: The url of the website to add.
        :return: 'Success' if the website was added to the knowledge base.
        """
        if self.knowledge is None:
            return "Knowledge base not provided"

        log_debug(f"Adding to knowledge base: {url}")
        self.knowledge.add_content(url=url)
        return "Success"

    def read_url(self, url: str) -> str:
        """This function reads a url and returns the content.

        :param url: The url of the website to read.
        :return: Relevant documents from the website.
        """
        from agno.knowledge.reader.website_reader import WebsiteReader

        website = WebsiteReader()

        log_debug(f"Reading website: {url}")
        relevant_docs: List[Document] = website.read(url=url)
        return json.dumps([doc.to_dict() for doc in relevant_docs])



from typing import Any, Optional

from agno.utils.log import logger


class AwsApiClient:
    def __init__(
        self,
        aws_region: Optional[str] = None,
        aws_profile: Optional[str] = None,
    ):
        super().__init__()
        self.aws_region: Optional[str] = aws_region
        self.aws_profile: Optional[str] = aws_profile

        # aws boto3 session
        self._boto3_session: Optional[Any] = None
        logger.debug("**-+-** AwsApiClient created")

    def create_boto3_session(self) -> Optional[Any]:
        """Create a boto3 session"""
        import boto3

        logger.debug("Creating boto3.Session")
        try:
            self._boto3_session = boto3.Session(
                region_name=self.aws_region,
                profile_name=self.aws_profile,
            )
            logger.debug("**-+-** boto3.Session created")
            logger.debug(f"\taws_region: {self._boto3_session.region_name}")
            logger.debug(f"\taws_profile: {self._boto3_session.profile_name}")
        except Exception as e:
            logger.error("Could not connect to aws. Please confirm aws cli is installed and configured")
            logger.error(e)
            exit(0)
        return self._boto3_session

    @property
    def boto3_session(self) -> Optional[Any]:
        if self._boto3_session is None:
            self._boto3_session = self.create_boto3_session()
        return self._boto3_session
